# Authors: Paul Kienzle, Brian Maranville
"""
Reflectometry backend loader
"""

import importlib

from . import BACKEND_NAME, BACKEND_NAMES

BACKEND_MODULE_NAMES = {
    "c_ext": "refl1d.reflmodule",
    "numba": "refl1d.lib.numba",
    "python": "refl1d.lib.python",
}

backend = None


def set_backend(backend_name: BACKEND_NAMES):
    global backend
    backend_module_name = BACKEND_MODULE_NAMES.get(backend_name, None)
    if backend_module_name is None:
        raise ValueError(f"Unknown backend: {backend_name}")
    backend = importlib.import_module(backend_module_name)


if BACKEND_NAME is not None:
    set_backend(BACKEND_NAME)



import wx
from bumps.gui.input_list import InputListPanel
from numpy import inf

from refl1d.names import SLD, FreeInterface, FreeLayer, PolymerBrush, Slab

# name, type, description, *args
# type: 'string'
# type: 'material' (see below)
# type: 'parameter' with units and limits (see below)
# type: '[parameter]' is a vector of parameters with common properties

# 'material' is a scatterer, which is one of:
#
#   SLD, with scattering length density rho + irho
#   Material, with compound and natural density, which can be fit
#      by density, natural density, relative density or cell volume.
#      The compound may be a chemical formula, or it may be multiple
#      formulas, with a count for each.
#   Mixture, which is a set of scatterers with a base making up the
#      bulk and % making up the rest.  The % is by mass or by volume.
#      If by mass, each material needs a density.  The scatterers can
#      be SLDs, materials or other mixtures.
#
# The material name defaults to the chemical formula, if available.
#
# Users must be able to select a material from another layer, with
# fittable parameters automatically tied to that other layer, and a
# visual indication that this is so.
#
# A database of scatterers used in various experiments would be useful.
#
# The periodictable class can be used to lookup expected SLD values
# as in the SANS SLD calculator.
#
# 'parameter' has name, value, and can be fixed, fitted or computed.
#
# Fitted parameters have a range, which can be value pm v, value pm v%,
# value pm (-v1,v2), value pm (-v1,+v2)%, or simply [v1,v2], where value
# is the current parameter value.  Maybe we can do this in the input
# box for value.
#
# Computed values need a formula, which can be any math expression,
# with a popup box to select parameter Pn from a table of Pn:name:path.
#
# Parameters may be copied directly from other fields with the same units,
# presumably using a combobox to select the parameter.
#
# Some of these operations will require more information than just the
# current stack.

# These belong with the layer definition, so that users can add their own
# layer types more easily.  I will use monkey patching until the details
# of the interface are worked out.
Slab.name = "slab"
Slab.description = "layer with uniform density and roughness on top"
Slab.fields = (
    ("name", "string", "slab name (defaults to material name)"),
    ("material", "material", "slab material"),
    ("thickness", "parameter", "slab thickness", "A", (0, inf)),
    ("interface", "parameter", "rms roughness between this slab and the next", "A", (0, inf)),
)

PolymerBrush.name = "brush"
PolymerBrush.description = "polymer brush in a solvent"
PolymerBrush.fields = (
    ("name", "string", "brush name (defaults to polymer name)"),
    ("polymer", "material", "polymer composition"),
    ("solvent", "material", "solvent composition"),
    ("thickness", "parameter", "brush+solvent thickness", "A", (0, inf)),
    ("interface", "parameter", "rms roughness between the solvent and the next layer", "A", (0, inf)),
    ("base_vf", "parameter", "volume fraction of the polymer brush at the base", "%", (0, 100)),
    ("base", "parameter", "thickness of the base region", "A", (0, inf)),
    ("length", "parameter", "thickness of the thinning region", "A", (0, inf)),
    ("power", "parameter", "rate of brush thinning", "", (-inf, inf)),
    ("sigma", "parameter", "rms roughness within the brush", "A", (0, inf)),
)

FreeLayer.name = "freeform"
FreeLayer.description = "freeform layer using monotonic splines for good control"
FreeLayer.fields = (
    ("name", "string", "layer name"),
    ("thickness", "string", "total layer thickness", "A", (0, inf)),
    # The following need to be links to neighbouring layer parameters; easy
    # enough with slabs, but somewhat messy with something like a tethered
    # polymer.  Perhaps we can define below and above fields for all the
    # layer types which act as scatterers to make this happen, but for now
    # just assume it is attached to a slab.  Insert/Delete are going to
    # have to update
    ("below", "material", "material below the layer"),
    ("above", "material", "material above the layer"),
    # Need to indicate that these are three coordinated vectors
    ("z", "[parameter]", "control point location", "A", (0, 1)),
    ("rho", "[parameter]", "scattering length density", "1e-6 inv A^2", (-inf, inf)),
    ("irho", "[parameter]", "complex scattering length density", "1e-6 inv A^2", (0, inf)),
)

FreeInterface.name = "blend"
FreeInterface.description = "blend between two materials using a monotonic spline"
FreeInterface.fields = (
    ("name", "string", "layer name"),
    ("below", "material", "material below the layer"),
    ("above", "material", "material above the layer"),
    ("thickness", "parameter", "total layer thickness", "A", (0, inf)),
    ("interface", "parameter", 'rms roughness between "above" and the next layer', "A", (0, inf)),
    # Need to indicate that these are two coordinated vectors
    ("dz", "[parameter]", "relative segment size", "", (0, inf)),
    ("dp", "[parameter]", "relative step height", "", (0, inf)),
)

LAYERS = (Slab, FreeInterface, FreeLayer, PolymerBrush)


class LayerEditorDialog(wx.Dialog):
    """
    Select layer type and edit layer parameters interactively.
    """

    def __init__(
        self,
        parent=None,
        id=wx.ID_ANY,  # @ReservedAssignment
        title="Layer Editor",
        pos=wx.DefaultPosition,
        size=(-1, 400),
        style=wx.DEFAULT_FRAME_STYLE,  # |wx.THICK_FRAME|wx.RESIZE_BORDER,
        name="",
        stack=None,
        layer_num=0,
    ):
        wx.Dialog.__init__(self, parent, id, title, pos, size, style, name)
        if parent is not None:
            self.SetFont(parent.GetFont())

        self.next_button = wx.Button(self, wx.ID_ANY, "Next", style=wx.BU_EXACTFIT)
        self.prev_button = wx.Button(self, wx.ID_ANY, "Prev", style=wx.BU_EXACTFIT)
        self.ins_button = wx.Button(self, wx.ID_ANY, "Insert", style=wx.BU_EXACTFIT)
        self.del_button = wx.Button(self, wx.ID_ANY, "Delete", style=wx.BU_EXACTFIT)

        self.layer_id = wx.StaticText(self, size=(-1, -1))
        self.selector = wx.Choice(self)
        self.fields = InputListPanel(self)

        # Generic bindings
        self.Bind(wx.EVT_BUTTON, self.OnNext, self.next_button)
        self.Bind(wx.EVT_BUTTON, self.OnPrev, self.prev_button)
        self.Bind(wx.EVT_BUTTON, self.OnInsert, self.ins_button)
        self.Bind(wx.EVT_BUTTON, self.OnDelete, self.del_button)
        self.Bind(wx.EVT_TEXT, self.OnValidate)
        # self.Bind(wx.EVT_COMBO_BOX, self.OnValidate)
        self.selector.Bind(wx.EVT_CHOICE, self.OnLayerType)

        # Layout
        outer = wx.BoxSizer(wx.VERTICAL)

        bbox = wx.GridSizer(rows=1, cols=4, vgap=0, hgap=5)
        bbox.Add(self.next_button, 1, wx.EXPAND)
        bbox.Add(self.prev_button, 1, wx.EXPAND)
        bbox.Add(self.ins_button, 1, wx.EXPAND)
        bbox.Add(self.del_button, 1, wx.EXPAND)
        outer.Add(bbox, 0, wx.ALIGN_LEFT | wx.LEFT | wx.RIGHT | wx.TOP, border=10)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        hbox.Add(self.layer_id, 0, wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL | wx.RIGHT, border=5)
        hbox.Add(self.selector, 0, wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)
        hbox.Add((0, 0), 1)
        outer.Add(hbox, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, border=10)
        outer.Add(self.fields, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, border=10)
        outer.Add((-1, 10))

        dx, dy = outer.GetMinSize()
        outer.SetMinSize((dx, dy + 100))
        outer.Fit(self)
        self.SetSizer(outer)

        # Set initial layer
        self.layer = None
        self.layer_num = layer_num
        self.set_stack(stack)

    def OnNext(self, event):
        self.set_layer(self.layer_num + 1)

    def OnPrev(self, event):
        self.set_layer(self.layer_num - 1)

    def OnDelete(self, event):
        del self.stack[self.layer_num]
        self.set_layer(self.layer_num)

    def OnInsert(self, event):
        slab = Slab(SLD(rho=1.0, name="layer"), thickness=100, interface=5)
        self.stack.insert(self.layer_num, slab)
        self.set_layer(self.layer_num)

    def OnLayerType(self, event):
        pass

    def set_stack(self, stack):
        if stack is None:
            return
        self.stack = stack
        self.set_layer(self.layer_num)

    def set_layer(self, layer_num):
        # Make sure the layer is valid
        N = len(self.stack)
        if layer_num < 0:
            layer_num = 0
        elif layer_num >= N:
            layer_num = N - 1
        self.layer_num = layer_num

        # Enable navigation and control buttons
        self.next_button.Enable(layer_num < N - 1)
        self.prev_button.Enable(layer_num > 0)
        self.del_button.Enable(N > 1)
        self.layer_id.SetLabel("Layer %d:" % layer_num)

        # Null case: no stack
        if N == 0:
            return  # Should never happen...

        # Update the layer contents
        if self.layer != self.stack[layer_num]:
            self.layer = self.stack[layer_num]
            self.draw_layer()

    def draw_layer(self):
        return
        if self.layer is None:
            return
        self.name_label = wx.StaticText(self, label="Name")
        self.name = wx.TextCtrl(self, wx.ID_ANY, size=(250, -1))

        self.thickness_label = wx.StaticText(self, label="Thickness")
        self.thickness = wx.TextCtrl(self, wx.ID_ANY, size=(100, -1))
        self.interface_label = wx.StaticText(self, label="Interface")
        self.interface = wx.TextCtrl(self, wx.ID_ANY, size=(100, -1))

        self.layer_name.SetValue(self.layer.name)
        self.thickness.SetValue("%g" % self.layer.thickness.value)
        self.interface.SetValue("%g" % self.layer.interface.value)
        # self.selector.SetValue(self.layer.__class__.__name__)

    def OnValidate(self, event):
        ctrl = event.GetEventObject()
        if ctrl:
            validator = ctrl.GetValidator()
            if validator:
                validator.Validate(ctrl)
        event.Skip()


def main(stack):
    class App(wx.App):
        def OnInit(self):
            dia = LayerEditorDialog(None, -1, "Mock-up")
            dia.set_stack(stack)
            dia.ShowModal()
            dia.Destroy()
            return True

    app = App(0)
    app.MainLoop()


if __name__ == "__main__":
    from refl1d.names import air, silicon

    stack = silicon | air
    main(stack)



# Below taken from the GUI profile plot methods in Refl1d - modified to be stand alone
# This is to allow for the layer labels to be easily added to the SLD plots

# Originally written for use in a jupyter notebook
# Adapted to act as a prototype for the profile plotting in the new GUI for refl1d

from typing import List, TypedDict

import numpy as np
from numpy import inf

from refl1d.experiment import Experiment

from .colors import COLORS as MULTI_PLOT_COLORS

SINGLE_PLOT_COLORS = ("black", "blue", "green", "gold")

SINGLE_PLOT_STYLES = ["solid"] * 4
MULTI_PLOT_STYLES = ["solid", "dash", "dashdot", "dot"]

# === Sample information ===


class FindLayers:
    def __init__(self, experiment, axes=None, x_offset=0):
        self.experiment = experiment
        self.axes = axes
        self.x_offset = x_offset
        self._show_labels = True
        self._show_boundaries = True

        self._find_layer_boundaries()

    def find(self, z):
        return self.experiment.sample.find(z - self.x_offset)

    def _find_layer_boundaries(self):
        offset = self.x_offset
        boundary = [-inf, offset]
        if hasattr(self.experiment, "sample"):
            for L in self.experiment.sample[1:-1]:
                dx = L.thickness.value
                offset += dx
                boundary.append(offset)
        boundary.append(inf)
        self.boundary = np.asarray(boundary)

    def sample_labels(self):
        return [L.name for L in self.experiment.sample]

    def sample_layer(self, n):
        return self.experiment.sample[n]

    def label_offsets(self):
        z = self.boundary[1:-1]
        left = -20
        middle = [(a + b) / 2.0 for a, b in zip(z[:-1], z[1:])]
        right = z[-1] + 20
        return [left] + middle + [right]

    def reset_markers_plotly(self):
        """
        Reset all markers, for plotly plots
        """
        import plotly.graph_objs as go

        # self.clear_markers()
        if not isinstance(self.axes, go.Figure):
            raise ValueError("reset_markers_plotly can only be used with type: axes=plotly.graph_objs.Figure")

        fig = self.axes

        # Add bars
        fittable = [self.experiment.sample[idx].thickness.fittable for idx, _ in enumerate(self.boundary[1:-1])]
        fittable[0] = False  # First interface is not fittable

        for f, z_pos in zip(fittable, self.boundary[1:-1]):
            if not f:
                line_dict = dict(dash="dash")
            else:
                line_dict = dict(dash="solid")

            fig.add_vline(
                x=z_pos,
                opacity=0.75,
                line=line_dict,
            )

        # for f,m in zip(fittable,self.markers):
        #     if not f: m.set(linestyle='--', linewidth=1.25)
        # self.connect_markers(m for f,m in zip(fittable,self.markers) if f)

        # Add labels
        offsets = self.label_offsets()
        labels = self.sample_labels()
        for label_pos, label in zip(offsets, labels):
            fig.add_annotation(
                text=label, textangle=-30, x=label_pos, yref="paper", y=1.0, yanchor="bottom", showarrow=False
            )


def generate_best_profile(model: Experiment):
    if model.ismagnetic:
        best = model.magnetic_smooth_profile(dz=0.5)
    else:
        best = model.smooth_profile()
    return best


# ============================================================================= #
# Plotting script below
# ============================================================================= #


class ModelSpec(TypedDict):
    model_index: int
    sample_index: int


class PlotItem(TypedDict):
    model: Experiment
    spec: ModelSpec
    color_index: int


def plot_multiple_sld_profiles(plot_items: List[PlotItem]):
    import plotly.graph_objs as go

    fig = go.Figure()
    nplots = len(plot_items)
    multiplot = nplots > 1
    for plot_item in plot_items:
        model = plot_item["model"]
        spec = plot_item["spec"]
        color_index = plot_item["color_index"]
        model_index = spec["model_index"]
        sample_index = spec["sample_index"]
        if multiplot:
            colors = [MULTI_PLOT_COLORS[color_index % len(MULTI_PLOT_COLORS)]] * 4
        else:
            colors = SINGLE_PLOT_COLORS
        line_styles = MULTI_PLOT_STYLES if multiplot else SINGLE_PLOT_STYLES
        legendgroup = f"{model_index}:{sample_index} ({model.name})" if multiplot else None
        plot_sld_profile_plotly(model, fig, colors=colors, line_styles=line_styles, legendgroup=legendgroup)
    return fig


def plot_sld_profile_plotly(model, fig, colors=SINGLE_PLOT_COLORS, line_styles=SINGLE_PLOT_STYLES, legendgroup=None):
    if model.ismagnetic:
        z_best, rho_best, irho_best, rhoM_best, thetaM_best = generate_best_profile(model)
        yaxis_title = "SLD: ρ, ρ<sub>i</sub>, ρ<sub>M</sub> / 10<sup>-6</sup> Å<sup>-2</sup>"
    else:
        z_best, rho_best, irho_best = generate_best_profile(model)
        yaxis_title = "SLD: ρ, ρ<sub>i</sub> / 10<sup>-6</sup> Å<sup>-2</sup>"

    hovertemplate = (
        f"(%{{x}}, %{{y}})<br>{legendgroup if legendgroup is not None else ''} %{{data.hovertext}}<extra></extra>"
    )

    fig.add_scatter(
        x=z_best,
        y=rho_best,
        name="ρ",
        legendgroup=legendgroup,
        legendgrouptitle_text=legendgroup,
        hovertemplate=hovertemplate,
        hovertext="SLD",
        line={
            "color": colors[0],
            "dash": line_styles[0],
        },
    )

    fig.add_scatter(
        x=z_best,
        y=irho_best,
        name="ρ<sub>i</sub>",
        legendgroup=legendgroup,
        hovertemplate=hovertemplate,
        hovertext="Im. SLD",
        line={
            "color": colors[1],
            "dash": line_styles[1],
        },
    )

    if model.ismagnetic:
        fig.add_scatter(
            x=z_best,
            y=rhoM_best,
            name="ρ<sub>M</sub>",
            legendgroup=legendgroup,
            hovertemplate=hovertemplate,
            hovertext="Mag. SLD",
            line={
                "color": colors[2],
                "dash": line_styles[2],
            },
        )

        fig.add_scatter(
            x=z_best,
            y=thetaM_best,
            name="θ<sub>M</sub>",
            yaxis="y2",
            legendgroup=legendgroup,
            hovertemplate=hovertemplate,
            hovertext="Theta M",
            line={
                "color": colors[3],
                "dash": line_styles[3],
            },
        )

        # TODO: need to make axis scaling for thetaM dependent on if thetaM exceeds 0-360
        fig.update_layout(
            yaxis2={
                "title": {"text": "Magnetic Angle θ<sub>M</sub> / °"},
                "type": "linear",
                "autorange": False,
                "range": [0, 360],
                "anchor": "x",
                "overlaying": "y",
                "side": "right",
                "ticks": "inside",
                # 'ticklen': 20,
            }
        )

    fig.update_layout(uirevision=1, plot_bgcolor="white")
    fig.update_layout(
        xaxis={
            "title": {"text": "depth (Å)"},
            "type": "linear",
            "autorange": True,
            # 'gridcolor': "Grey",
            "ticks": "inside",
            # 'ticklen': 20,
            "showline": True,
            "linewidth": 1,
            "linecolor": "black",
            "mirror": "ticks",
            "side": "bottom",
        }
    )

    fig.update_layout(
        yaxis={
            "title": {"text": yaxis_title},
            "exponentformat": "e",
            "showexponent": "all",
            "type": "linear",
            "autorange": True,
            # 'gridcolor': "Grey",
            "ticks": "inside",
            # 'ticklen': 20,
            "showline": True,
            "linewidth": 1,
            "linecolor": "black",
            "mirror": True,
        }
    )

    fig.update_layout(margin={"l": 75, "r": 50, "t": 50, "b": 75, "pad": 4})

    fig.update_layout(
        legend={
            "x": -0.1,
            "bgcolor": "rgba(255,215,0,0.15)",
            # "traceorder": "reversed"
        }
    )

    marker_positions = FindLayers(model, axes=fig)
    marker_positions.reset_markers_plotly()

    # fig.show(renderer='firefox')
    return fig


# def plot_contours(model, title, savepath=None, nsamples=1000, show_contours=True, show_mag=True, savetitle=None,
#                   store=None, ultranest=False, dream=False, align=0):
#     if show_contours:
#         # data, columns = generate_contour_data(model)
#         if ultranest:
#             points_array = model.post_samples
#             sub_array = points_array[np.sort(np.random.randint(points_array.shape[0], size=nsamples)), :]
#             errors = calc_errors(model.problem, sub_array)
#             data, columns = generate_profile_data(errors)
#         # if this comment is removed the above code will not overwrite the previous reference to the errors object
#         # and it will act as a bad memory leak - using more ram everytime the function is called
#         if dream:
#             state, points = load_dream_state_notebook(problem=model.problem, store=store)
#             errors = calc_errors(model.problem, points[-nsamples:-1])
#             data, columns = generate_profile_data(errors, align)
#
#     # get best profile manually as we are not using DREAM state
#     if show_mag:
#         z_best, rho_best, irho_best, rhoM_best, thetaM_best = generate_best_profile(model)
#     else:
#         z_best, rho_best, irho_best = generate_best_profile(model)
#
#     fig, ax = plt.subplots(1, 1, figsize=(8, 6))
#
#     # rho
#     if show_contours:
#         z, best, sig2lo, sig2hi, siglo, sighi = data[0]
#         ax.plot(z, best, label=r"$\rho$ SLD", color="orange")
#         ax.fill_between(z, siglo, sighi, alpha=0.5, color="orange", label=r"$\rho$ SLD $1\sigma$")
#         ax.fill_between(z, sig2lo, sig2hi, alpha=0.25, color="orange", label=r"$\rho$ SLD $2\sigma$")
#     else:
#         ax.plot(z_best, rho_best, label=r"$\rho$ SLD", color="orange")
#
#     # irho
#     # ax.plot(z_best, irho_best, label=r"$\rho_{i}$ Im SLD")
#     # if show_contours:
#     #     z, best, sig2lo, sig2hi, siglo, sighi = data[1]
#     #     ax.fill_between(z, siglo, sighi, alpha=0.5)
#     #     ax.fill_between(z, sig2lo, sig2hi, alpha=0.25)
#     if show_mag:
#         if show_contours:
#             z, best, sig2lo, sig2hi, siglo, sighi = data[2]
#             ax.plot(z, best, label=r"$\rho_{M}$ mSLD", color="b")
#             ax.fill_between(z, siglo, sighi, alpha=0.5, color="b", label=r"$\rho_{M}$ mSLD $1\sigma$")
#             ax.fill_between(z, sig2lo, sig2hi, alpha=0.25, color="b", label=r"$\rho_{M}$ mSLD $2\sigma$")
#         else:
#             ax.plot(z_best, rhoM_best, label=r"$\rho_{M}$ mSLD", color="b")
#
#     # ax.legend(fontsize=14, loc='upper right', framealpha=0.5, ncol=2)
#     ax.legend(fontsize=16, framealpha=0.5, ncol=3)
#
#     ax.grid(True)
#     ax.set_xlabel(r"z $\left(\AA\right)$", fontsize=18)
#     ax.set_ylabel(r"$\rho, \rho_{i} \left(10^{-6}\AA^{-2}\right)$", fontsize=18)
#     if show_mag:
#         ax.set_ylabel(r"$\rho, \rho_{M} \left(10^{-6}\AA^{-2}\right)$", fontsize=18)
#     ax.tick_params(axis='y', labelsize=16)
#     ax.tick_params(axis='x', labelsize=16)
#
#     experiment = model.determine_problem_fitness()
#
#     layer_markers = FindLayers(experiment, axes=ax)
#     layer_markers.reset_markers()
#     fig.suptitle(title)
#     # if np.max(rho_best) > 75:
#     #     top = np.max(rho_best)*1.15
#     # else:
#     #     top =75
#     # ax.set_ylim(top=top)
#     if savepath:
#         if savetitle:
#             title = savetitle
#         plt.savefig(f"{savepath}/{title}.png")



# Author: Paul Kienzle
r"""
Scattering length density profile.

In order to render a reflectometry model, the theory function calculator
renders each layer in the model for each energy in the probe.  For slab
layers this is easy: just accumulate the slabs, with the 1-\ $\sigma$ Gaussian
interface width between the slabs.  For freeform or functional layers,
this is more complicated.  The rendering needs to chop each layer into
microslabs and evaluate the profile at each of these slabs.

Example
-------

This example sets up a model which uses tanh to transition from
silicon to gold in 20 |Ang| with 2 |Ang| steps.

First define the profile, and put in the substrate:

    >>> S = Microslabs(nprobe=1, dz=2)
    >>> S.clear()
    >>> S.append(w=0, rho=2.07)

Next add the interface.  This uses :meth:`microslabs` to select
the points at which the interface is evaluated, much like you
would do when defining your own special layer type.  Note that the
points Pz are in the center of the micro slabs.  The width of the
final slab may be different.  You do not need to use fixed width
microslabs if you can more efficiently represent the profile with
a smaller number of variable width slabs, but :meth:`contract_profile`
serves the same purpose with less work on your part.

    >>> from numpy import tanh
    >>> Pw, Pz = S.microslabs(20)
    >>> print("widths = %s ..."%(" ".join("%g"%v for v in Pw[:5])))
    widths = 2 2 2 2 2 ...
    >>> print("centers = %s ..."%(" ".join("%g"%v for v in Pz[:5])))
    centers = 1 3 5 7 9 ...
    >>> rho = (1-tanh((Pz-10)/5))/2*(2.07-4.5)+4.5
    >>> S.extend(w=Pw, rho=[rho])

Finally, add the incident medium and see the results.  Note that *rho*
is a matrix, with one column for each incident energy.  We are only
using one energy so we only show the first column.

    >>> S.append(w=0, rho=4.5)
    >>> print("width = %s ..."%(" ".join("%g"%v for v in S.w[:5])))
    width = 0 2 2 2 2 ...
    >>> print("rho = %s ..."%(" ".join("%.2f"%v for v in S.rho[0, :5])))
    rho = 2.07 2.13 2.21 2.36 2.63 ...

 Since *irho* and *sigma* were not specified, they will be zero.

    >>> print("sigma = %s ..."%(" ".join("%g"%v for v in S.sigma[:5])))
    sigma = 0 0 0 0 0 ...
    >>> print("irho = %s ..."%(" ".join("%g"%v for v in S.irho[0, :5])))
    irho = 0 0 0 0 0 ...
"""

import numpy as np
from numpy import isnan, nan
from scipy.special import erf

from .sample.reflectivity import BASE_GUIDE_ANGLE as DEFAULT_THETA_M


class Microslabs(object):
    """
    Manage the micro slab representation of a model.

    In order to compute reflectivity, we need a series of slabs with thickness,
    roughness and scattering potential for each slab.  Because scattering
    potentials are probe dependent we store an array of potentials for each
    probe value.

    Some slab models use non-uniform layers, and so need the additional
    parameter of dz for the step size within the layer.

    The space for the slabs is saved even after reset, in preparation for a
    new set of slabs from different fitting parameters.

    """

    def __init__(self, nprobe, dz=1):
        self._num_slabs = 0
        # _slabs contains the 1D objects w, sigma of len n
        # _slabs_rho contains 2D objects rho, irho, with one for each wavelength
        # rhoM, thetaM will contain magnetic moment and angle
        self._slabs = np.empty(shape=(0, 2))
        self._slabs_rho = np.empty(shape=(0, nprobe, 2))
        self.rhoM = None  # type: np.ndarray
        self.thetaM = None  # type: np.ndarray
        self._slabs_mag = np.empty(shape=(0, nprobe, 2))
        self.dz = dz
        self._magnetic_sections = []
        self._z_left = self._z_right = 0.0
        self._z_offset = 0.0

    def microslabs(self, thickness=0):
        """
        Return a set of microslabs for a layer of the given *thickness*.

        The step size slabs.dz was defined when the Microslabs
        object was created.

        This is a convenience function.  Layer definitions can choose
        their own slices so long as the step size is approximately
        slabs.dz in the varying region.

        :Parameters:
            *thickness* : float | A
                Layer thickness
        :Returns:
            *widths*: vector | A
                Microslab widths
            *centers*: vector | A
                Microslab centers
        """
        # TODO: force dz onto a common boundary to avoid remeshing
        # in the smooth profile function
        edges = np.arange(0, thickness + self.dz, self.dz, dtype="d")
        edges[-1] = thickness
        centers = (edges[1:] + edges[:-1]) / 2
        widths = edges[1:] - edges[:-1]
        return widths, centers

    def clear(self):
        """
        Reset the slab model so that none are present.
        """
        self._num_slabs = 0
        self._magnetic_sections = []

    def __len__(self):
        return self._num_slabs

    def repeat(self, start=0, count=1, interface=0):
        """
        Extend the model so that there are *count* versions of the slabs
        from *start* to the final slab.

        This is equivalent to L.extend(L[start:]*(count-1)) for list L.
        """
        # For now use the dumb implementation; a better implementation
        # would remember the repeats and pre-calculate the matrix product
        # for the repeating region, saving much work later.  This has
        # to work in conjunction with interfaces and with magnetic profiles.
        repeats = count - 1
        end = len(self)
        length = end - start
        fromidx = slice(start, end)
        toidx = slice(end, end + repeats * length)
        self._reserve(repeats * length)
        self._slabs[toidx] = np.tile(self._slabs[fromidx], [repeats, 1])
        self._slabs_rho[toidx] = np.tile(self._slabs_rho[fromidx], [repeats, 1, 1])
        self._num_slabs += repeats * length

        # Replace interface on the top
        self._slabs[self._num_slabs - 1, 1] = interface

        if self._magnetic_sections:
            raise NotImplementedError("Repeated magnetic layers not implemented")

    def _reserve(self, nadd):
        """
        Reserve space for at least *nadd* slabs.
        """
        ns, nl, _ = self._slabs_rho.shape
        if ns < self._num_slabs + nadd:
            new_ns = self._num_slabs + nadd + 50
            self._slabs = self._slabs.copy()
            self._slabs.resize((new_ns, 4), refcheck=False)
            self._slabs_rho = self._slabs_rho.copy()
            self._slabs_rho.resize((new_ns, nl, 2), refcheck=False)

    def extend(self, w=0, sigma=0, rho=0, irho=0):
        """
        Extend the micro slab model with the given layers.
        """
        nadd = len(w)
        self._reserve(nadd)
        idx = slice(self._num_slabs, self._num_slabs + nadd)
        self._num_slabs += nadd
        self._slabs[idx, 0] = w
        self._slabs[idx, 1] = sigma
        self._slabs_rho[idx, :, 0] = np.asarray(rho).T
        self._slabs_rho[idx, :, 1] = np.asarray(irho).T

    def append(self, w=0, sigma=0, rho=0, irho=0):
        """
        Extend the micro slab model with a single layer.
        """
        # self.extend(w=[w], sigma=[sigma], rho=[rho], irho=[irho])
        # return
        self._reserve(1)
        self._slabs[self._num_slabs, 0] = w
        self._slabs[self._num_slabs, 1] = sigma
        self._slabs_rho[self._num_slabs, :, 0] = rho
        self._slabs_rho[self._num_slabs, :, 1] = irho
        self._num_slabs += 1

    def add_magnetism(self, anchor, w, rhoM=0, thetaM=DEFAULT_THETA_M, sigma=0):
        """
        Add magnetic layers.

        Note that *sigma* is a pair *(interface_below, interface_above)*
        representing the magnetic roughness, which may be different
        from the nuclear roughness at the layer boundaries.
        """
        w = np.asarray(w, "d")
        if np.isscalar(sigma):
            sigma = (sigma, sigma)
        self._magnetic_sections.append((np.vstack((w, rhoM, thetaM)), anchor, sigma))

    def thickness(self):
        """
        Total thickness of the profile.

        Note that thickness includes the thickness of the substrate and
        surface layers.  Normally these will be zero, but the contract
        profile operation may result in large values for either.
        """
        return np.sum(self._slabs[1 : self._num_slabs, 0])

    @property
    def w(self):
        "Thickness (A)"
        return self._slabs[: self._num_slabs, 0]

    @property
    def sigma(self):
        "rms roughness (A)"
        return self._slabs[: self._num_slabs - 1, 1]

    @property
    def surface_sigma(self):
        "roughness for the current top layer, or nan if substrate"
        return self._slabs[self._num_slabs - 1, 1] if self._num_slabs > 0 else nan

    @property
    def rho(self):
        "Scattering length density (10^-6 number density)"
        return self._slabs_rho[: self._num_slabs, :, 0].T

    @property
    def irho(self):
        "Absorption (10^-6 number density)"
        return self._slabs_rho[: self._num_slabs, :, 1].T

    @property
    def ismagnetic(self):
        "True if there are magnetic materials in any slab"
        return self._magnetic_sections != []

    def limited_sigma(self, limit=0):
        """
        Limit the roughness by some fraction of layer thickness.

        This function should be called before :meth:`finalize`, but after
        all slabs have been added to the profile.

        *limit* is the number of times sigma has to fit in the layers
        on either side of the interface.  The returned sigma is
        truncated to min(wlow, whigh)/*limit* where wlow is the thickness
        of the layer below the interface, and whigh is the thickness above
        the interface.  A *limit* value of 0 returns the original sigma.
        Although a gaussian inteface extends to infinity, in practice
        setting a *limit* of 3 allows the layer to reach its bulk value,
        with no cross talk between the interfaces.  For very large
        roughnesses, the blending algorithm allows the sld beyond
        the interface to bleed through the entire layer and into the
        next.  In this case the roughness should be the same on both
        sides of the layer to avoid artifacts at the interface.

        Magnetic roughness is ignored for now.
        """
        self.sigma[:] = compute_limited_sigma(self.w, self.sigma, limit)

    def finalize(self, step_interfaces, dA):
        """
        Rendering complete.

        Call this method after the microslab model has been constructed,
        so any post-rendering processes can be completed.

        In addition to clearing any width from the substrate and
        the surface surround, this will align magnetic and nuclear slabs,
        convert interfaces to step interfaces if desired, and merge slabs
        with similar scattering potentials to reduce computation time.

        *step_interfaces* is True if interfaces should be rendered using
        slabs.

        *dA* is the tolerance to use when deciding if similar layers can
        be merged.
        """
        if self.ismagnetic:
            self._align_magnetic_and_nuclear()

        self._set_z_range()

        # render step interfaces
        if step_interfaces:
            self._render_interfaces()

        if self.ismagnetic:
            self._contract_magnetic(dA)
        else:
            self._contract_profile(dA)

    def _set_z_range(self):
        """
        Make sure z-range includes 3-sigma around every interface.
        """
        self.w[0] = self.w[-1] = 0
        offset = np.cumsum(self.w[:-1])
        self._z_left = min(-10, np.min(offset - 3 * self.sigma))
        self._z_right = max(offset[-1] + 10, np.max(offset + 3 * self.sigma))

    def _align_magnetic_and_nuclear(self):
        """
        Add magnetic information to the nuclear slabs, introducing new
        slabs as necessary where magnetic and nuclear do not match.
        """
        from .backends import backend

        # Nuclear profile (one wavelength only)
        # if self.rho.shape[0] != 1:
        #    raise ValueError("wavelength-dependent magnetism not supported")
        w, sigma, rho, irho = self.w, self.sigma, self.rho[0], self.irho[0]

        # Fill in gaps for magnetic profile
        wM, sigmaM, rhoM, thetaM = self._join_magnetic_sections(gap_size=1e-6)

        # Align nuclear and magnetic
        w, sigma, rho, irho, wM, sigmaM, rhoM, thetaM = [
            np.ascontiguousarray(v, "d") for v in (w, sigma, rho, irho, wM, sigmaM, rhoM, thetaM)
        ]
        output = np.empty((len(w) + len(wM), 6), "d")
        n = backend.align_magnetic(w, sigma, rho, irho, wM, sigmaM, rhoM, thetaM, output)

        # Store the resulting profile
        self._reserve(n - self._num_slabs)  # make sure there is space
        self._num_slabs = n
        self.w[:] = output[:n, 0]
        self.sigma[:] = output[: n - 1, 1]
        self.rho[0][:] = output[:n, 2]
        self.irho[0][:] = output[:n, 3]
        self.rhoM = output[:n, 4]
        self.thetaM = output[:n, 5]

    def _render_interfaces(self):
        """
        Convert all interfaces into step interfaces by sampling the
        analytic version of the smoothed profile at intervals of dz.

        The interface effects are limited to the surrounding layers.

        Use of contract_profile afterward is strongly recommended, for
        better performance on models with large sections of constant
        scattering potential.
        """
        z = np.arange(self._z_left, self._z_right + 0.5 * self.dz, self.dz)
        n_slabs = len(z)
        n_profiles = self.rho.shape[0]
        offsets = np.cumsum(self.w[:-1])  # assumes w[0] == 0 in _set_z_range

        # generate profiles
        Nrho = len(self.rho)
        Nirho = len(self.irho)
        # assume that irho has the same shape...
        to_stack = [self.rho, self.irho]
        if self.ismagnetic:
            to_stack.extend([self.rhoM[None, :], self.thetaM[None, :]])

        value = np.vstack(to_stack)

        profiles = _build_profiles_backend(z, offsets, self.sigma, value)

        rho = profiles[0:Nrho]
        # print('rho:', self.rho.shape, rho.shape)
        irho = profiles[Nrho : Nrho + Nirho]
        if self.ismagnetic:
            rhoM = profiles[Nrho + Nirho]
            thetaM = profiles[Nrho + Nirho + 1]

        w = self.dz * np.ones(n_slabs)
        w[0] = w[-1] = 0.0

        # update slabs
        self._reserve(n_slabs - self._num_slabs)
        self._num_slabs = n_slabs
        self.w[:] = w
        self.sigma[:] = 0
        self.rho[:] = rho
        self.irho[:] = irho
        if self.ismagnetic:
            self.rhoM = rhoM
            self.thetaM = thetaM
        self._z_offset = self._z_left

    def _contract_profile(self, dA):
        from .backends import backend

        if dA is None:
            return

        # TODO: need a separate implementation for multiple wavelengths
        if self.rho.shape[0] > 1:
            # Note: should at least check for duplicates otherwise thick
            # layers will get extremely slow
            return

        w, sigma, rho, irho = [np.ascontiguousarray(v, "d") for v in (self.w, self.sigma, self.rho[0], self.irho[0])]
        # print "final sld before contract", rho[-1]
        n = backend.contract_by_area(w, sigma, rho, irho, dA)
        self._num_slabs = n
        self.w[:] = w[:n]
        self.rho[0, :] = rho[:n]
        self.irho[0, :] = irho[:n]
        self.sigma[:] = sigma[: n - 1]
        # print "final sld after contract", rho[n-1], self.rho[0][n-1], n

    def _contract_magnetic(self, dA):
        from .backends import backend

        if dA is None:
            return

        # TODO: need a separate implementation for multiple wavelengths
        if self.rho.shape[0] > 1:
            # Note: should at least check for duplicates otherwise thick
            # layers will get extremely slow
            return

        w, sigma, rho, irho, rhoM, thetaM = [
            np.ascontiguousarray(v, "d")
            for v in (self.w, self.sigma, self.rho[0], self.irho[0], self.rhoM, self.thetaM)
        ]
        # print "final sld before contract", rho[-1]
        n = backend.contract_mag(w, sigma, rho, irho, rhoM, thetaM, dA)
        self._num_slabs = n
        self.w[:] = w[:n]
        self.rho[0][:] = rho[:n]
        self.irho[0][:] = irho[:n]
        self.rhoM = rhoM[:n]
        self.thetaM = thetaM[:n]
        self.sigma[:] = sigma[: n - 1]
        # self.sigma[:] = 0
        # print "final sld after contract", rho[n-1], self.rho[0][n-1], n

    def _DEAD_apply_smoothness(self, dA, smoothness=0.3):
        """
        Set a guassian interface for layers which have been coalesced using
        the contract_profile function.

        Note that we guess which interfaces this applies to after the fact
        using criteria similar to those used to coalesce the microslabs
        into layers, and so it may apply to layers which are close in
        scattering length density and have zero sigma, but which were
        distinct in the original model.  The displayed profile will show
        the profile used to calculate the reflectivity, so even though
        this behaviour is different from what the user intended, the
        result will not be misleading.

        In a detailed example of a tethered polymer model, smoothness
        was found to be worse than no smoothness, so this function has
        been removed from the execution stream.
        """
        if dA is None or smoothness == 0:
            return
        # TODO: refine this so that it can look forward as well as back
        # TODO: also need to avoid changing explicit sigma=0...
        w, sigma, rho, irho = self.w, self.sigma, self.rho[0], self.irho[0]
        step = abs(np.diff(rho)) + abs(np.diff(irho))
        step[:-1] *= w[1:-1]  # compute dA of step; substrate uses w=1
        idx = np.nonzero(sigma == 0)[0]
        fix = step[idx] < 3 * dA
        sigma[idx[fix]] = w[idx[fix]] * smoothness

    def step_profile(self):
        """
        Return a step profile representation of the microslab structure.

        Nevot-Croce interfaces are not represented.
        """
        rho = np.vstack([self.rho[0, :], self.rho[0, :]]).T.flatten()
        irho = np.vstack([self.irho[0, :], self.irho[0, :]]).T.flatten()
        if len(self.w) > 2:
            offsets = np.cumsum(self.w[0:-1])
            z = np.vstack([np.hstack([-10, offsets]), np.hstack([offsets, offsets[-1] + 10])]).T.flatten()
        else:
            z = np.array([-10, 0, 0, 10])
        return z + self._z_offset, rho, irho

    def magnetic_step_profile(self):
        """
        Return a step profile representation of the microslab structure.

        Nevot-Croce interfaces are not represented.
        """
        z, rho, irho = self.step_profile()
        rhoM = np.vstack([self.rhoM, self.rhoM]).T.flatten()
        thetaM = np.vstack([self.thetaM, self.thetaM]).T.flatten()
        return z + self._z_offset, rho, irho, rhoM, thetaM

    def smooth_profile(self, dz=0.1):
        """
        Return a smooth profile representation of the microslab structure

        Nevot-Croce roughness is approximately represented, though the
        calculation is incorrect for layers with large roughness compared
        to the thickness.

        The returned profile has uniform step size *dz*.
        """
        z = np.arange(self._z_left, self._z_right + 0.5 * dz, dz)
        offsets = np.cumsum(self.w[:-1]) + self._z_offset
        values = np.vstack([self.rho[0], self.irho[0]])
        profiles = _build_profiles_backend(z, offsets, self.sigma, values)
        rho, irho = profiles
        return z, rho, irho

    def magnetic_smooth_profile(self, dz=0.1):
        """
        Return a profile representation of the magnetic microslab structure.
        """
        z = np.arange(self._z_left, self._z_right + 0.5 * dz, dz)
        offsets = np.cumsum(self.w[:-1]) + self._z_offset
        values = np.vstack([self.rho[0], self.irho[0], self.rhoM, self.thetaM])
        profiles = _build_profiles_backend(z, offsets, self.sigma, values)
        rho, irho, rhoM, thetaM = profiles
        return z, rho, irho, rhoM, thetaM

    def _join_magnetic_sections(self, gap_size):
        """
        Convert anchored magnetic sections into coarse magnetic slabs.
        """
        # Find the magnetic blocks
        blocks, offsets, sigmas = zip(*self._magnetic_sections)

        # print "blocks", blocks
        # print "offsets", offsets
        # print "sigmas", sigmas
        # * Splice the blocks together with rhoM=0 in the gaps and
        #   thetaM=(thetaM_below+thetaM_above)/2 in the gaps.
        # * Result is:
        #   slices = [(thickness, rhoM, thetaM), (thickness, rhoM, thetaM), ...]
        # * Initialize slices with the magnetism of the substrate, which will
        #   be [thickness=0, rhoM=0, thetaM=first thetaM] unless substrate
        #   magnetism has been specified
        substrate_magnetism = isnan(sigmas[0][0])
        if substrate_magnetism:
            slices = [[[], [], []]]
        else:
            slices = [[[0], [0], [blocks[0][2, 0]]]]
        interfaces = []
        pos = 0
        for i, B in enumerate(blocks):
            anchor = offsets[i]
            w = anchor - pos
            if w >= gap_size:  # Big gap, so need spacer
                # Target average theta between blocks.
                if i == 0:
                    thetaM = B[2, 0]
                    interfaces.append(0)
                else:
                    thetaM = (B[2, 0] + blocks[i - 1][2, -1]) / 2.0
                    interfaces.append(sigmas[i - 1][1])
                slices.append([[w], [0], [thetaM]])
                interfaces.append(sigmas[i][0])
            elif w >= -1e-6:
                # Small gap, so add it to the start of the next block
                B[0, 0] += w
                anchor -= w
                if i == 0:
                    if not substrate_magnetism:
                        interfaces.append(sigmas[0][0])
                else:
                    # Use interface_above between blocks which are connected,
                    # ignoring interface_below.
                    interfaces.append(sigmas[i - 1][1])
            else:
                # negative gap should never happen
                raise ValueError("Overlapping magnetic layers at %d" % i)
            slices.append(B)
            nslabs = len(B[0, :])
            interfaces.extend([0] * (nslabs - 1))
            width = np.sum(B[0, :])
            pos = anchor + width

        # Add the final slice
        w = self.thickness() - pos
        theta = blocks[-1][2, -1]
        slices.append([[w], [0], [theta]])
        interfaces.append(sigmas[-1][1])

        wM, rhoM, thetaM = [np.hstack(v) for v in zip(*slices)]
        sigmaM = np.array(interfaces)
        # print "result", wM, rhoM, thetaM, sigmaM
        return wM, sigmaM, rhoM, thetaM


def compute_limited_sigma(thickness, roughness, limit):
    # Limit roughness to the depths of the surrounding layers.  Roughness
    # of the first and last layers interfaces is limited only by the
    # depth of the first and last layers.  We must check explicitly for
    # a pure substrate system since that has no limits on roughness.
    if limit > 0 and len(thickness) > 2:
        s = np.min((thickness[:-1], thickness[1:]), axis=0) / limit
        s[+0] = thickness[+1] / limit
        s[-1] = thickness[-2] / limit
        roughness = np.where(roughness < s, roughness, s)
    return roughness


def _build_profiles_backend(z, offsets, roughness, value):
    from .backends import backend

    contrast = (value[:, 1:] - value[:, :-1]).ravel(order="C")
    initial_value = value[:, 0].copy()  # contiguous

    NZ = len(z)
    # Number of profiles:
    NP = initial_value.shape[0]
    profiles = np.zeros((NP, NZ), dtype=float).ravel("C")
    backend.build_profile(z, offsets.copy(), roughness.copy(), contrast, initial_value, profiles)
    return profiles.reshape((NP, NZ))


def build_profile(z, offset, roughness, value):
    """
    Convert a step profile to a smooth profile.

    *z*          calculation points
    *offset*     offset for each interface
    *roughness*  roughness of each interface
    *value*      target value for each slab
    *max_rough*  limit the roughness to a fraction of the layer thickness
    """
    contrast = np.diff(value)
    result = np.zeros_like(z) + value[0]
    for offset_k, sigma_k, contrast_k in zip(offset, roughness, contrast):
        delta = contrast_k * blend(z, sigma_k, offset_k)
        result += delta
    return result


SQRT1_2 = 1.0 / np.sqrt(2.0)


def blend(z, sigma, offset):
    """
    blend function

    Given a Gaussian roughness value, compute the portion of the neighboring
    profile you expect to find in the current profile at depth z.
    """
    if sigma <= 0.0:
        return 1.0 * (z >= offset)
    else:
        return 0.5 * erf(SQRT1_2 * (z - offset) / sigma) + 0.5



# This program is in the public domain
# Author: Andrew Nelson
"""
ANSTO data loaders

The following instrument is defined::

    Platypus

All the ANSTO instruments emit Q/R/dR/dQ in their output files.
"""

import os.path
import re

import numpy as np
from bumps.data import maybe_open

from ..probe import QProbe
from ..resolution import FWHM2sigma


def _load_dat(f):
    """
    Loads a Platypus dataset from file. This will normally be Q, R, dR, dQ.

    | Q - Momentum transfer |1/Ang|
    | R - Reflectivity
    | dR - uncertainty in reflectivity (1 sigma)
    | dQ - FWHM of Gaussian resolution kernel.

    **Parameters**

    f : file-handle or string
        File to load the dataset from.

    **Returns**

    (filename, name, data) - str, str, tuple of np.ndarray
    """

    # it would be nice to use refl1d.probe.data_loaders.load4 directly. However,
    # there are lots of files in the wild that don't use # to denote comments
    # in the header. In addition, it is known that ANSTO datasets emit dQ as
    # FWHM.

    with maybe_open(f) as fh:
        fname = fh.name
        header_lines = 0
        for i, line in enumerate(fh):
            try:
                nums = [float(tok) for tok in re.split(r"\s|,", line) if len(tok)]
            except ValueError:
                continue
            if len(nums) >= 2:
                header_lines = i
                break

    data = np.loadtxt(f, unpack=True, skiprows=header_lines)

    filename = fname
    name = os.path.splitext(os.path.basename(fname))[0]

    return filename, name, data


def load(filename, instrument=None, **kw):
    """
    Return a probe for ANSTO data.

    **Parameters**

    f : file-handle or string
        File to load the dataset from.

    **Returns**

    probe : probe.QProbe
    """
    fname, name, data = _load_dat(filename)

    Q = data[0]
    R = data[1]
    dR = data[2]
    dQ = FWHM2sigma(data[3])

    probe = QProbe(Q, dQ, data=(R, dR), name=name, filename=fname)

    if instrument is not None:
        probe.instrument = instrument.instrument

    return probe


class ANSTOData(object):
    def load(self, filename, **kw):
        return load(filename, instrument=self, **kw)


class Platypus(ANSTOData):
    """
    Loader for reduced data from the ANSTO Platypus instrument.
    """

    instrument = "Platypus"
    radiation = "neutron"


INSTRUMENTS = {
    "Platypus": Platypus,
}



import numpy as np

from .config import pick_radius
from .config import profile_color
from .config import vf_scale
from .interactor import BaseInteractor
from .layer import MaterialInteractor
from .util import clip, setpar


# -----------------------------------------------------------------------
class FreeInterfaceInteractor(BaseInteractor):
    """
    Interactor for freeform interfaces using monotonic splines.
    """

    def __init__(self, profile, layer):
        super(FreeInterfaceInteractor, self).__init__(profile)
        self.below = MaterialInteractor(profile, layer.below, range=(0, 0.3))
        self.above = MaterialInteractor(profile, layer.above, range=(0.7, 1))
        self.layer = layer
        ax = profile.axes

        style = dict(
            linestyle=":",
            transform=profile.xcoords,
            zorder=0,
            color=profile_color,
            visible=True,
        )
        self.hprofile = ax.plot([], [], **style)[0]

        style = dict(
            marker="o",
            transform=profile.xcoords,
            zorder=7,
            pickradius=pick_radius,
            color=profile_color,
            alpha=0.5,
            markersize=5,
            visible=True,
        )
        self.markers = [ax.plot([], [], **style)[0] for _ in layer.dz[:-1]]
        self.connect_markers(self.markers)

    def update_markers(self):
        """
        Draw the widgets in their new positions.
        """
        self.below.update_markers()
        self.above.update_markers()

        n = self.profile.layer_num
        left = self.profile.boundary[n]
        layer = self.layer

        z = np.linspace(0, layer.thickness.value, 200)
        vf = layer.profile(z)
        self.hprofile.set_xdata(z + left)
        self.hprofile.set_ydata(vf * vf_scale)

        z = np.cumsum(np.array([v.value for v in layer.dz], "d"))
        p = np.cumsum(np.array([v.value for v in layer.dp], "d"))
        self._zscale = layer.thickness.value / z[-1]
        if p[-1] == 0:
            p[-1] = 1
        p *= vf_scale / p[-1]
        z *= self._zscale
        z += left
        for h, zi, pi in zip(self.markers, z[:-1], p[:-1]):
            h.set_data((zi,), (pi,))
        # self.markers[0].set_data(z,p)

    def clear_markers(self):
        self.below.clear_markers()
        self.above.clear_markers()
        super(FreeInterfaceInteractor, self).clear_markers()
        self.hprofile.remove()

    def drag(self, evt):
        """
        Update the model with the new widget position.
        """
        n = self.profile.layer_num
        left, right = self.profile.boundary[n : n + 2]

        idx = self.markers.index(evt.artist)
        z = [h.get_xdata()[0] for h in self.markers]
        z[idx] = evt.xdata
        dz = np.diff(np.sort(np.clip(z + [left, right], left, right)))
        dz /= np.max(dz)
        for p, v in zip(self.layer.dz, dz):
            setpar(p, v)

    def save(self, evt):
        """
        Save the current state of the model represented by the widget.
        """
        self._save_dz = [p.value for p in self.layer.dz]
        self._save_dp = [p.value for p in self.layer.dp]

    def restore(self, evt):
        """
        Restore the widget and model to the saved state.
        """
        for p, v in zip(self.layer.dz, self._save_dz):
            p.value = v
        for p, v in zip(self.layer.dp, self._save_dp):
            p.value = v


# -----------------------------------------------------------------------
class FreeLayerInteractor(BaseInteractor):
    """
    Interactor for freeform layers using monotonic splines.
    """

    def __init__(self, profile, layer):
        super(FreeLayerInteractor, self).__init__(profile)
        self.below = MaterialInteractor(profile, layer.below, range=(0, 0.3))
        self.above = MaterialInteractor(profile, layer.above, range=(0.7, 1))
        self.layer = layer
        ax = profile.axes

        style = dict(
            marker="o",
            zorder=7,
            pickradius=pick_radius,
            color=profile_color,
            alpha=0.5,
            markersize=5,
            visible=True,
        )
        self.markers = [ax.plot([], [], **style)[0] for _ in layer.rho + layer.irho]
        self.connect_markers(self.markers)

    def update_markers(self):
        """
        Draw the widgets in their new positions.
        """
        self.below.update_markers()
        self.above.update_markers()

        n = self.profile.layer_num
        left = self.profile.boundary[n]
        layer = self.layer

        z = np.array([v.value for v in layer.z])
        rho = np.array([v.value for v in layer.rho])
        rhoI = np.array([v.value for v in layer.irho])
        self._zscale = layer.thickness.value
        z *= self._zscale
        z += left
        for h, zi, pi in zip(self.markers[: len(rho)], z, rho):
            h.set_data((zi,), (pi,))
        for h, zi, pi in zip(self.markers[len(rho) :], z, rhoI):
            h.set_data((zi,), (pi,))

    def clear_markers(self):
        self.below.clear_markers()
        self.above.clear_markers()
        super(FreeLayerInteractor, self).clear_markers()

    def drag(self, evt):
        """
        Update the model with the new widget position.
        """
        n = self.profile.layer_num
        left, right = self.profile.boundary[n : n + 2]
        z, rho, rhoI = self.layer.z, self.layer.rho, self.layer.irho
        nrho = len(rho)

        zval = clip((evt.xdata - left) / (right - left), 0, 1)
        idx = self.markers.index(evt.artist)
        if idx >= nrho:
            idx -= nrho
            setpar(z[idx], zval)
            setpar(rhoI[idx], evt.ydata)
        else:
            setpar(z[idx], zval)
            setpar(rho[idx], evt.ydata)

    def save(self, evt):
        """
        Save the current state of the model represented by the widget.
        """
        self._save_z = [p.value for p in self.layer.z]
        self._save_rho = [p.value for p in self.layer.rho]
        self._save_rhoI = [p.value for p in self.layer.irho]

    def restore(self, evt):
        """
        Restore the widget and model to the saved state.
        """
        for p, v in zip(self.layer.z, self._save_z):
            p.value = v
        for p, v in zip(self.layer.rho, self._save_rho):
            p.value = v
        for p, v in zip(self.layer.irho, self._save_rhoI):
            p.value = v



__all__ = ["merge_ends"]

import numpy as np


def asbytes(s):
    return s.encode("utf-8")


def merge_ends(w, p, tol=1e-3):
    """
    join the leading and trailing ends of the profile together so fewer
    slabs are required and so that gaussian roughness can be used.
    """
    # TODO: accept rho, rhoi pairs as well
    # TODO: make sure we apply an interface to the right as well as the left
    try:
        # Assuming there p[0] != p[-1] within tolerance, we are guaranteed
        # that we will have a first value not equivalent to p[0] on the
        # left, with index > 0 and a last value not equivalent to p[-1] on
        # the right, with index < -1.  We are going to put the first value
        # at left index - 1 and the last value at right index + 1, accumulating
        # the widths of the identical layers.
        lidx = np.where(abs(p - p[0]) > tol)[0][0] - 1
        ridx = len(p) - np.where(abs(p[::-1] - p[-1]) > tol)[0][0]
        w[lidx], p[lidx] = np.sum(w[: lidx + 1]), p[0]
        w[ridx], p[ridx] = np.sum(w[ridx:]), p[-1]
        return w[lidx : ridx + 1], p[lidx : ridx + 1]
    except Exception:
        if len(w):
            # All one big layer
            w[0] = np.sum(w)
            return w[0:1], p[0:1]
        else:
            return w, p



# from bumps.webview.server.api import state, set_problem, load_session
# from .webserver import main, start_app, create_server_task, sio


from bumps.webview.server.api import state, set_problem, load_session
from bumps.webview.server.webserver import start_app, create_server_task, sio

from .webserver import main



__all__ = [
    "reflectivity_amplitude",
    "magnetic_amplitude",
    "calculate_u1_u3",
    "build_profile",
    "convolve_gaussian",
    "convolve_uniform",
    "convolve_sampled",
    "align_magnetic",
    "contract_by_area",
    "contract_mag",
    "rebin_counts",
    "rebin_counts_2D",
]

from .reflectivity import reflectivity_amplitude
from .magnetic import magnetic_amplitude
from .magnetic import calculate_u1_u3
from .build_profile import build_profile
from .convolve import convolve_gaussian
from .convolve import convolve_uniform
from .convolve_sampled import convolve_sampled
from .contract_profile import align_magnetic
from .contract_profile import contract_by_area
from .contract_profile import contract_mag
from .rebin import rebin_counts
from .rebin import rebin_counts_2D



#!/usr/bin/env python
# This program is in the public domain.
# Authors: Paul Kienzle and James Krycka
"""
Start Refl1D with the command line interface.
"""

# ========================== Start program ==================================
# Process the command line that has been entered.
if __name__ == "__main__":
    # This is necessary when running the application from a frozen image and
    # using the --parallel option.  Note that freeze_support() has no effect
    # when running from a python script (i.e., in a non-frozen environment).
    import multiprocessing

    multiprocessing.freeze_support()

    import refl1d.main

    refl1d.main.cli()



# -*- coding: utf-8 -*-
# Authors Paul Kienzle, Richard Sheridan

r"""
Layer models for polymer systems.

Analytic Self-consistent Field (SCF) Brush profile\ [#Zhulina]_\ [#Karim]_

Analytical Self-consistent Field (SCF) Mushroom Profile\ [#Adamuti-Trache]_

Numerical Self-consistent Field (SCF) End-Tethered Polymer
Profile\ [#Cosgrove]_\ [#deVos]_\ [#Sheridan]_


.. [#Zhulina] Zhulina, EB; Borisov, OV; Pryamitsyn, VA; Birshtein, TM (1991)
    "Coil-Globule Type Transitions in Polymers. 1. Collapse of Layers
    of Grafted Polymer Chains", Macromolecules 24, 140-149.

.. [#Karim] Karim, A; Douglas, JF; Horkay, F; Fetters, LJ; Satija, SK (1996)
    "Comparative swelling of gels and polymer brush layers",
    Physica B 221, 331-336. doi:10.1016/0921-4526(95)00946-9

.. [#Adamuti-Trache] Adamuţi-Trache, M., McMullen, W. E. & Douglas, J. F.
    Segmental concentration profiles of end-tethered polymers with
    excluded-volume and surface interactions. J. Chem. Phys. 105, 4798 (1996).

.. [#Cosgrove] Cosgrove, T., Heath, T., Van Lent, B., Leermakers, F. A. M.,
    & Scheutjens, J. M. H. M. (1987). Configuration of terminally attached
    chains at the solid/solvent interface: self-consistent field theory and
    a Monte Carlo model. Macromolecules, 20(7), 1692–1696.
    doi:10.1021/ma00173a041

.. [#deVos] De Vos, W. M., & Leermakers, F. A. M. (2009). Modeling the
    structure of a polydisperse polymer brush. Polymer, 50(1), 305–316.
    doi:10.1016/j.polymer.2008.10.025

.. [#Sheridan] Sheridan, R. J., Orski, S. V., Jones, R. L., Satija, S.,
    & Beers, K. L. (2017). Surface interaction parameter measurement of
    solvated polymers via model end-tethered chains. [Submitted]

..  [#Vincent] Vincent, B., Edwards, J., Emmett, S., & Croot, R. (1988).
    Phase separation in dispersions of weakly-interacting particles in
    solutions of non-adsorbing polymer. Colloids and Surfaces, 31, 267–298.
    doi:10.1016/0166-6622(88)80200-2

"""

__all__ = ["PolymerBrush", "PolymerMushroom", "EndTetheredPolymer", "VolumeProfile", "layer_thickness"]

import inspect
from collections import OrderedDict
from time import time

from bumps.parameter import Parameter
import numpy as np
from numpy import exp, hstack, imag, log, ones_like, pi, real, sqrt

try:
    from numpy._core.multiarray import correlate as old_correlate
except ImportError:
    # cruft for numpy < 2
    from numpy.core.multiarray import correlate as old_correlate

from .. import utils
from .layers import Layer

LAMBDA_1 = 1.0 / 6.0  # always assume cubic lattice (1/6) for now
LAMBDA_0 = 1.0 - 2.0 * LAMBDA_1
# Use reverse order for LAMBDA_ARRAY if it is asymmetric since we are using
# it with correlate().
LAMBDA_ARRAY = np.array([LAMBDA_1, LAMBDA_0, LAMBDA_1])
MINLAT = 25
MINBULK = 5
SQRT_PI = sqrt(pi)


class PolymerBrush(Layer):
    r"""
    Polymer brushes in a solvent

    :Parameters:
        *thickness*
            the thickness of the solvent layer
        *interface*
            the roughness of the solvent surface
        *polymer*
            the polymer material
        *solvent*
            the solvent material or vacuum
        *base_vf*
            volume fraction (%) of the polymer brush at the interface
        *base*
            the thickness of the brush interface (A)
        *length*
            the length of the brush above the interface (A)
        *power*
            the rate of brush thinning
        *sigma*
            rms brush roughness (A)

    The materials can either use the scattering length density directly,
    such as PDMS = SLD(0.063, 0.00006) or they can use chemical composition
    and material density such as PDMS=Material("C2H6OSi", density=0.965).

    These parameters combine in the following profile formula:

    .. math::

        V(z) &= \left\{
          \begin{array}{ll}
            V_o                        & \mbox{if } z <= z_o \\
            V_o (1 - ((z-z_o)/L)^2)^p  & \mbox{if } z_o < z < z_o + L \\
            0                          & \mbox{if } z >= z_o + L
          \end{array}
        \right. \\
        V_\sigma(z)
           &= V(z) \star
                 \frac{e^{-\frac{1}{2}(z/\sigma)^2}}{\sqrt{2\pi\sigma^2}} \\
        \rho(z) &= \rho_p V_\sigma(z) + \rho_s (1-V_\sigma(z))

    where $V_\sigma(z)$ is volume fraction convoluted with brush
    roughness $\sigma$ and $\rho(z)$ is the complex scattering
    length density of the profile.
    """

    def __init__(
        self,
        thickness=0,
        interface=0,
        name="brush",
        polymer=None,
        solvent=None,
        base_vf=None,
        base=None,
        length=None,
        power=None,
        sigma=None,
    ):
        prefix = name + " "
        self.thickness = Parameter.default(thickness, name=prefix + "thickness")
        self.interface = Parameter.default(interface, name=prefix + "interface")
        self.base_vf = Parameter.default(base_vf, name=prefix + "base_vf")
        self.base = Parameter.default(base, name=prefix + "base")
        self.length = Parameter.default(length, name=prefix + "length")
        self.power = Parameter.default(power, name=prefix + "power")
        self.sigma = Parameter.default(sigma, name=prefix + "sigma")
        self.solvent = solvent
        self.polymer = polymer
        self.name = name
        # Constraints:
        #   base_vf in [0, 1]
        #   base, length, sigma, thickness, interface > 0
        #   base + length + 3*sigma <= thickness

    def parameters(self):
        return {
            "solvent": self.solvent.parameters(),
            "polymer": self.polymer.parameters(),
            "base_vf": self.base_vf,
            "base": self.base,
            "length": self.length,
            "power": self.power,
            "sigma": self.sigma,
        }

    def profile(self, z):
        base_vf, base, length, power, sigma = [
            p.value for p in (self.base_vf, self.base, self.length, self.power, self.sigma)
        ]
        base_vf /= 100.0  # % to fraction
        L0 = base  # if base < thickness else thickness
        L1 = base + length  # if base+length < thickness else thickness-L0
        if length == 0:
            v = np.ones_like(z)
        else:
            v = 1 - ((z - L0) / (L1 - L0)) ** 2
        v[z < L0] = 1
        v[z > L1] = 0
        brush_profile = base_vf * v**power
        # TODO: we could use Nevot-Croce rather than smearing the profile
        vf = smear(z, brush_profile, sigma)
        return vf

    def render(self, probe, slabs):
        thickness = self.thickness.value
        Pw, Pz = slabs.microslabs(thickness)
        # Skip layer if it falls to zero thickness.  This may lead to
        # problems in the fitter, since R(thickness) is non-differentiable
        # at thickness = 0.  "Clip to boundary" range handling will at
        # least allow this point to be found.
        # TODO: consider using this behaviour on all layer types.
        if len(Pw) == 0:
            return

        Mr, Mi = self.polymer.sld(probe)
        Sr, Si = self.solvent.sld(probe)
        M = Mr + 1j * Mi
        S = Sr + 1j * Si
        try:
            M, S = M[0], S[0]  # Temporary hack
        except Exception:
            pass

        vf = self.profile(Pz)
        Pw, vf = utils.merge_ends(Pw, vf, tol=1e-3)
        P = M * vf + S * (1 - vf)
        Pr, Pi = real(P), imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)


def layer_thickness(z):
    """
    Return the thickness of a layer given the microslab z points.

    The z points are at the centers of the bins.  we can use the recurrence
    that boundary b[k] = z[k-1] + (z[k-1] - b[k-1]) to compute the
    total length of the layer.
    """
    return 2 * (np.sum(z[-1::-2]) - np.sum(z[-2::-2]))


class VolumeProfile(Layer):
    """
    Generic volume profile function

    :Parameters:

        *thickness*
            the thickness of the solvent layer
        *interface*
            the roughness of the solvent surface
        *material*
            the polymer material
        *solvent*
            the solvent material
        *profile*
            the profile function, suitably parameterized

    The materials can either use the scattering length density directly,
    such as PDMS = SLD(0.063, 0.00006) or they can use chemical composition
    and material density such as PDMS=Material("C2H6OSi", density=0.965).

    These parameters combine in the following profile formula::

        sld = material.sld * profile + solvent.sld * (1 - profile)

    The profile function takes a depth z and returns a density rho.

    For volume profiles, the returned rho should be the volume fraction
    of the material.  For SLD profiles, rho should be complex scattering
    length density of the material.

    Fitting parameters are the available named arguments to the function.
    The first argument must be *z*, which is the array of depths at which
    the profile is to be evaluated.  It is guaranteed to be increasing, with
    step size 2*z[0].

    Initial values for the function parameters can be given using name=value.
    These values can be scalars or fitting parameters.  The function will
    be called with the current parameter values as arguments.  The layer
    thickness can be computed as :func: `layer_thickness`.

    """

    # TODO: test that thickness(z) matches the thickness of the layer
    def __init__(self, thickness=0, interface=0, name="VolumeProfile", material=None, solvent=None, profile=None, **kw):
        if interface != 0:
            raise NotImplementedError("interface not yet supported")
        if profile is None or material is None or solvent is None:
            raise TypeError("Need polymer, solvent and profile")
        self.name = name
        self.thickness = Parameter.default(thickness, name="solvent thickness")
        self.interface = Parameter.default(interface, name="solvent interface")
        self.profile = profile
        self.solvent = solvent
        self.material = material

        # Query profile function for the list of arguments
        vars = inspect.getfullargspec(profile)[0]
        # print("vars", vars)
        if inspect.ismethod(profile):
            vars = vars[1:]  # Chop self
        vars = vars[1:]  # Chop z
        # print(vars)
        unused = [k for k in kw.keys() if k not in vars]
        if len(unused) > 0:
            raise TypeError("Profile got unexpected keyword argument '%s'" % unused[0])
        dups = [k for k in vars if k in ("thickness", "interface", "polymer", "solvent", "profile")]
        if len(dups) > 0:
            raise TypeError("Profile has conflicting argument '%s'" % dups[0])
        for k in vars:
            kw.setdefault(k, 0)
        for k, v in kw.items():
            setattr(self, k, Parameter.default(v, name=k))

        self._parameters = vars

    def parameters(self):
        P = {
            "solvent": self.solvent.parameters(),
            "material": self.material.parameters(),
        }
        for k in self._parameters:
            P[k] = getattr(self, k)
        return P

    def render(self, probe, slabs):
        Mr, Mi = self.material.sld(probe)
        Sr, Si = self.solvent.sld(probe)
        M = Mr + 1j * Mi
        S = Sr + 1j * Si
        # M, S = M[0], S[0]  # Temporary hack
        Pw, Pz = slabs.microslabs(self.thickness.value)
        if len(Pw) == 0:
            return
        kw = dict((k, getattr(self, k).value) for k in self._parameters)
        # print(kw)
        phi = self.profile(Pz, **kw)
        try:
            if phi.shape != Pz.shape:
                raise Exception
        except Exception:
            raise TypeError("profile function '%s' did not return array phi(z)" % self.profile.__name__)
        Pw, phi = utils.merge_ends(Pw, phi, tol=1e-3)
        P = M * phi + S * (1 - phi)
        slabs.extend(rho=[real(P)], irho=[imag(P)], w=Pw)
        # slabs.interface(self.interface.value)


def smear(z, P, sigma):
    """
    Gaussian smearing

    :Parameters:
        *z* | vector
            equally spaced sample times
        *P* | vector
            sample values
        *sigma* | real
            root-mean-squared convolution width
    :Returns:
        *Ps* | vector
            smeared sample values
    """
    if len(z) < 3:
        return P
    dz = z[1] - z[0]
    if 3 * sigma < dz:
        return P
    w = int(3 * sigma / dz)
    G = exp(-0.5 * (np.arange(-w, w + 1) * (dz / sigma)) ** 2)
    full = np.hstack(([P[0]] * w, P, [P[-1]] * w))
    return np.convolve(full, G / np.sum(G), "valid")


class PolymerMushroom(Layer):
    r"""
    Polymer mushrooms in a solvent (volume profile)

    :Parameters:
        *delta* | real scalar
            interaction parameter
        *vf* | real scalar
            not quite volume fraction (dimensionless grafting density)
        *sigma* | real scalar
            convolution roughness (A)

    Using analytical SCF methods for gaussian chains, which are scaled
    by the radius of gyration of the equivalent free polymer as an
    approximation to results of renormalization group methods.\ [#Adamuti-Trache]_

    Solutions are only strictly valid for vf << 1.
    """

    def __init__(self, thickness=0, interface=0, name="Mushroom", polymer=None, solvent=None, sigma=0, vf=0, delta=0):
        self.thickness = Parameter.default(thickness, name="Mushroom thickness")
        self.interface = Parameter.default(interface, name="Mushroom interface")
        self.delta = Parameter.default(delta, name="delta")
        self.vf = Parameter.default(vf, name="vf")
        self.sigma = Parameter.default(sigma, name="sigma")
        self.solvent = solvent
        self.polymer = polymer
        self.name = name

    def parameters(self):
        return {
            "solvent": self.solvent.parameters(),
            "polymer": self.polymer.parameters(),
            "delta": self.delta,
            "vf": self.vf,
            "sigma": self.sigma,
            "thickness": self.thickness,
            "interface": self.interface,
        }

    def profile(self, z):
        delta, sigma, vf, thickness = [p.value for p in (self.delta, self.sigma, self.vf, self.thickness)]

        return smear(z, MushroomProfile(z, delta, vf, sigma), sigma)

    def render(self, probe, slabs):
        thickness = self.thickness.value
        Pw, Pz = slabs.microslabs(thickness)
        # Skip layer if it falls to zero thickness.  This may lead to
        # problems in the fitter, since R(thickness) is non-differentiable
        # at thickness = 0.  "Clip to boundary" range handling will at
        # least allow this point to be found.
        # TODO: consider using this behaviour on all layer types.
        if len(Pw) == 0:
            return

        Mr, Mi = self.polymer.sld(probe)
        Sr, Si = self.solvent.sld(probe)
        M = Mr + 1j * Mi
        S = Sr + 1j * Si
        try:
            # TODO: Fix this hack
            M, S = M[0], S[0]  # Temporary hack
        except Exception:
            pass

        phi = self.profile(Pz)
        Pw, phi = utils.merge_ends(Pw, phi, tol=1e-3)
        P = M * phi + S * (1 - phi)
        Pr, Pi = np.real(P), np.imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)


def MushroomProfile(z, delta=0.1, vf=1.0, sigma=1.0):
    thickness = layer_thickness(z)
    thresh = 1e-10
    base = 3.0 * sigma  # tail is erf, capture 95% of the mixing
    Rg = (thickness - base) / 4.0  # profile ends by ~4 RG, so we can tether these
    keep = (z - base) >= 0.0
    x = (z[keep] - base) / Rg

    """
    mushroom_profile_math has a divide by zero problem at delta=0.
    Fix it by weighted average of the profile above and below a threshold.
    No visual difference when delta is between +-0.001, and there's no
    floating point error until ~+-1e-14.
    """
    if abs(delta) > thresh:
        mushroom_profile = mushroom_math(x, delta, vf)
    else:  # we should RARELY get here
        scale = (delta + thresh) / 2.0 / thresh
        mushroom_profile = scale * mushroom_math(x, thresh, vf) + (1.0 - scale) * mushroom_math(x, -thresh, vf)

    try:
        # make the base connect with the profile
        zextra = z[np.logical_not(keep)]
        base_profile = ones_like(zextra) * mushroom_profile[0]
    except IndexError:
        base_profile = ones_like(z) * mushroom_profile[0]

    return hstack((base_profile, mushroom_profile))


def mushroom_math(x, delta=0.1, vf=0.1):
    """
    new method, rewrite for numerical stability at high delta
    delta=0 causes divide by zero error!! Compensate elsewhere.
    http://ab-initio.mit.edu/wiki/index.php/Faddeeva_Package
    """
    from scipy.special import erfc, erfcx

    x_half = x / 2.0
    delta_double = 2.0 * delta
    return (
        (
            erfc(x_half)
            - erfcx(delta_double + x_half) / exp(x_half * x_half)
            - erfc(x)
            + ((0.25 - delta * (x + delta_double)) * erfcx(delta_double + x) + delta / SQRT_PI) * 4.0 / exp(x * x)
        )
        * vf
        / (delta_double * erfcx(delta_double))
    )


class EndTetheredPolymer(Layer):
    r"""
    Polymer end-tethered to an interface in a solvent

    Uses a numerical self-consistent field profile.\ [#Cosgrove]_\ [#deVos]_\ [#Sheridan]_

    **Parameters**
        *chi*
            solvent interaction parameter
        *chi_s*
            surface interaction parameter
        *h_dry*
            thickness of the neat polymer layer
        *l_lat*
            real length per lattice site
        *mn*
            Number average molecular weight
        *m_lat*
            real mass per lattice segment
        *pdi*
            Dispersity (Polydispersity index)
        *phi_b*
            volume fraction of free chains in solution. useful for associating
            grafted films e.g. PS-COOH in Toluene with an SiO2 surface.
        *thickness*
            Slab thickness should be greater than the contour
            length of the polymer
        *interface*
            should be zero
        *material*
            the polymer material
        *solvent*
            the solvent material

    Previous layer should not have roughness! Use a spline to simulate it.

    According to [#Vincent]_, $l_\text{lat}$ and $m_\text{lat}$ should be
    calculated by the formulas:

    .. math::

        l_\text{lat} &= \frac{a^2 m/l}{p_l} \\
        m_\text{lat} &= \frac{(a m/l)^2}{p_l}

    where $l$ is the real polymer's bond length, $m$ is the real segment mass,
    and $a$ is the ratio between molecular weight and radius of gyration at
    theta conditions. The lattice persistence, $p_l$, is:

    .. math::

        p_l = \frac16 \frac{1+1/Z}{1-1/Z}

    with coordination number $Z = 6$ for a cubic lattice, $p_l = .233$.
    """

    def __init__(
        self,
        thickness=0,
        interface=0,
        name="EndTetheredPolymer",
        polymer=None,
        solvent=None,
        chi=0,
        chi_s=0,
        h_dry=None,
        l_lat=1,
        mn=None,
        m_lat=1,
        pdi=1,
        phi_b=0,
    ):
        if interface != 0:
            raise NotImplementedError("interface not yet supported")
        if polymer is None or solvent is None or h_dry is None or mn is None:
            raise TypeError("Need polymer, solvent and profile")

        self.thickness = Parameter.default(thickness, name="SCF thickness")
        self.interface = Parameter.default(interface, name="SCF interface")
        self.chi = Parameter.default(chi, name="Chi")
        self.chi_s = Parameter.default(chi_s, name="Surface chi")
        self.h_dry = Parameter.default(h_dry, name="Dry thickness")
        self.l_lat = Parameter.default(l_lat, name="Lattice layer length")
        self.mn = Parameter.default(mn, name="Num. avg. MW")
        self.m_lat = Parameter.default(m_lat, name="Lattice segegment mass")
        self.pdi = Parameter.default(pdi, name="Dispersity")
        self.phi_b = Parameter.default(phi_b, name="Free polymer conc.")
        self.solvent = solvent
        self.polymer = polymer
        self.name = name

    def parameters(self):
        return {
            "solvent": self.solvent.parameters(),
            "polymer": self.polymer.parameters(),
            "chi": self.chi,
            "chi_s": self.chi_s,
            "h_dry": self.h_dry,
            "l_lat": self.l_lat,
            "mn": self.mn,
            "m_lat": self.m_lat,
            "pdi": self.pdi,
            "phi_b": self.phi_b,
            "thickness": self.thickness,
            "interface": self.interface,
        }

    def profile(self, z):
        return SCFprofile(
            z,
            chi=self.chi.value,
            chi_s=self.chi_s.value,
            h_dry=self.h_dry.value,
            l_lat=self.l_lat.value,
            mn=self.mn.value,
            m_lat=self.m_lat.value,
            pdi=self.pdi.value,
            phi_b=self.phi_b.value,
        )

    def render(self, probe, slabs):
        thickness = self.thickness.value
        Pw, Pz = slabs.microslabs(thickness)
        # Skip layer if it falls to zero thickness.  This may lead to
        # problems in the fitter, since R(thickness) is non-differentiable
        # at thickness = 0.  "Clip to boundary" range handling will at
        # least allow this point to be found.
        # TODO: consider using this behaviour on all layer types.
        if len(Pw) == 0:
            return

        Mr, Mi = self.polymer.sld(probe)
        Sr, Si = self.solvent.sld(probe)
        M = Mr + 1j * Mi
        S = Sr + 1j * Si
        try:
            M, S = M[0], S[0]  # Temporary hack
        except Exception:
            pass

        phi = self.profile(Pz)
        Pw, phi = utils.merge_ends(Pw, phi, tol=1e-3)
        P = M * phi + S * (1 - phi)
        Pr, Pi = np.real(P), np.imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)


def SCFprofile(z, chi=None, chi_s=None, h_dry=None, l_lat=1, mn=None, m_lat=1, phi_b=0, pdi=1, disp=False):
    """
    Generate volume fraction profile for Refl1D based on real parameters.

    The field theory is a lattice-based one, so we need to move between lattice
    and real space. This is done using the parameters l_lat and m_lat, the
    lattice size and the mass of a lattice segment, respectivley. We use h_dry
    (dry thickness) as a convenient measure of surface coverage, along with mn
    (number average molecular weight) as the real inputs.

    Make sure your inputs for h_dry/l_lat and mn/m_lat match dimensions!
    Angstroms and daltons are good choices.

    This function is suitable for use as a VolumeProfile, as well as the
    default EndTetheredPolymer class.
    """

    # calculate lattice space parameters
    theta = h_dry / l_lat
    segments = mn / m_lat
    sigma = theta / segments

    # solve the self consistent field equations using the cache
    if disp:
        print("\n=====Begin calculations=====\n")
    phi_lat = SCFcache(chi, chi_s, pdi, sigma, phi_b, segments, disp)
    if disp:
        print("\n============================\n")

    # Chop edge effects out
    for x, layer in enumerate(reversed(phi_lat)):
        if abs(layer - phi_b) < 1e-6:
            break
    phi_lat = phi_lat[: -(x + 1)]

    # re-dimensionalize the solution
    layers = len(phi_lat)
    z_end = l_lat * layers
    z_lat = np.linspace(0.0, z_end, num=layers)
    phi = np.interp(z, z_lat, phi_lat, right=phi_b)

    return phi


_SCFcache_dict = OrderedDict()


def SCFcache(chi, chi_s, pdi, sigma, phi_b, segments, disp=False, cache=_SCFcache_dict):
    """Return a memoized SCF result by walking from a previous solution.

    Using an OrderedDict because I want to prune keys FIFO
    """
    try:
        from scipy.optimize import NoConvergence
    except ImportError:
        # cruft for scipy < 1.14, hard breaking change with no warning
        from scipy.optimize.nonlin import NoConvergence
    # prime the cache with a known easy solutions
    if not cache:
        cache[(0, 0, 0, 0.1, 0.1, 0.1)] = SCFsolve(sigma=0.1, phi_b=0.1, segments=50, disp=disp)
        cache[(0, 0, 0, 0, 0.1, 0.1)] = SCFsolve(sigma=0, phi_b=0.1, segments=50, disp=disp)
        cache[(0, 0, 0, 0.1, 0, 0.1)] = SCFsolve(sigma=0.1, phi_b=0, segments=50, disp=disp)

    if disp:
        starttime = time()

    # Try to keep the parameters between 0 and 1. Factors are arbitrary.
    scaled_parameters = (chi, chi_s * 3, pdi - 1, sigma, phi_b, segments / 500)

    # longshot, but return a cached result if we hit it
    if scaled_parameters in cache:
        if disp:
            print("SCFcache hit at:", scaled_parameters)
        phi = cache[scaled_parameters] = cache.pop(scaled_parameters)
        return phi

    # Find the closest parameters in the cache: O(len(cache))

    # Numpy setup
    cached_parameters = tuple(dict.__iter__(cache))
    cp_array = np.array(cached_parameters)
    p_array = np.array(scaled_parameters)

    # Calculate distances to all cached parameters
    deltas = p_array - cp_array  # Parameter space displacement vectors
    closest_index = np.sum(deltas * deltas, axis=1).argmin()

    # Organize closest point data for later use
    closest_cp = cached_parameters[closest_index]
    closest_cp_array = cp_array[closest_index]
    closest_delta = deltas[closest_index]

    phi = cache[closest_cp] = cache.pop(closest_cp)

    if disp:
        print("Walking from nearest:", closest_cp_array)
        print("to:", p_array)

    """
    We must walk from the previously cached point to the desired region.
    This is goes from step=0 (cached) and step=1 (finish), where the step=0
    is implicit above. We try the full step first, so that this function only
    calls SCFsolve one time during normal cache misses.

    The solver may not converge if the step size is too big. In that case,
    we retry with half the step size. This should find the edge of the basin
    of attraction for the solution eventually. On successful steps we increase
    stepsize slightly to accelerate after getting stuck.

    It might seem to make sense to bin parameters into a coarser grid, so we
    would be more likely to have cache hits and use them, but this rarely
    happened in practice.
    """

    step = 1.0  # Fractional distance between cached and requested
    dstep = 1.0  # Step size increment
    flag = True

    while flag:
        # end on 1.0 exactly every time
        if step >= 1.0:
            step = 1.0
            flag = False

        # conditional math because, "why risk floating point error"
        if flag:
            p_tup = tuple(closest_cp_array + step * closest_delta)
        else:
            p_tup = scaled_parameters

        if disp:
            print("Parameter step is", step)
            print("current parameters:", p_tup)

        try:
            phi = SCFsolve(
                p_tup[0], p_tup[1] / 3, p_tup[2] + 1, p_tup[3], p_tup[4], p_tup[5] * 500, disp=disp, phi0=phi
            )
        except (NoConvergence, ValueError) as e:
            if isinstance(e, ValueError):
                if str(e) != "array must not contain infs or NaNs":
                    raise
            if disp:
                print("Step failed")
            flag = True  # Reset this so we don't quit if step=1.0 fails
            dstep *= 0.5
            step -= dstep
            if dstep < 1e-5:
                raise RuntimeError("Cache walk appears to be stuck")
        else:  # Belongs to try, executes if no exception is raised
            cache[p_tup] = phi
            dstep *= 1.05
            step += dstep

    if disp:
        print("SCFcache execution time:", round(time() - starttime, 3), "s")

    # keep the cache from consuming all things
    while len(cache) > 100:
        cache.popitem(last=False)

    return phi


def SCFsolve(chi=0, chi_s=0, pdi=1, sigma=None, phi_b=0, segments=None, disp=False, phi0=None, maxiter=30):
    """Solve SCF equations using an initial guess and lattice parameters

    This function finds a solution for the equations where the lattice size
    is sufficiently large.

    The Newton-Krylov solver really makes this one. With gmres, it was faster
    than the other solvers by quite a lot.
    """

    from scipy.optimize import newton_krylov

    if sigma >= 1:
        raise ValueError("Chains that short cannot be squeezed that high")

    if disp:
        starttime = time()

    p_i = SZdist(pdi, segments)

    if phi0 is None:
        # TODO: Better initial guess for chi>.6
        phi0 = default_guess(segments, sigma)
        if disp:
            print("No guess passed, using default phi0: layers =", len(phi0))
    else:
        phi0 = abs(phi0)
        phi0[phi0 > 0.99999] = 0.99999
        if disp:
            print("Initial guess passed: layers =", len(phi0))

    # resizing loop variables
    jac_solve_method = "gmres"
    lattice_too_small = True

    # We tolerate up to 1 ppm deviation from bulk phi
    # when counting layers_near_phi_b
    tol = 1e-6

    def curried_SCFeqns(phi):
        return SCFeqns(phi, chi, chi_s, sigma, segments, p_i, phi_b)

    while lattice_too_small:
        if disp:
            print("Solving SCF equations")

        try:
            with np.errstate(invalid="ignore"):
                phi = abs(
                    newton_krylov(
                        curried_SCFeqns,
                        phi0,
                        verbose=bool(disp),
                        maxiter=maxiter,
                        method=jac_solve_method,
                    )
                )
        except RuntimeError as e:
            if str(e) == "gmres is not re-entrant":
                # Threads are racing to use gmres. Lose the race and use
                # something slower but thread-safe.
                jac_solve_method = "lgmres"
                continue
            else:
                raise

        if disp:
            print("lattice size:", len(phi))

        phi_deviation = abs(phi - phi_b)
        layers_near_phi_b = phi_deviation < tol
        nbulk = np.sum(layers_near_phi_b)
        lattice_too_small = nbulk < MINBULK

        if lattice_too_small:
            # if there aren't enough layers_near_phi_b, grow the lattice 20%
            newlayers = max(1, round(len(phi0) * 0.2))
            if disp:
                print("Growing undersized lattice by", newlayers)
            if nbulk:
                i = np.diff(layers_near_phi_b).nonzero()[0].max()
            else:
                i = phi_deviation.argmin()
            phi0 = np.insert(phi, i, np.linspace(phi[i - 1], phi[i], num=newlayers))

    if nbulk > 2 * MINBULK:
        chop_end = np.diff(layers_near_phi_b).nonzero()[0].max()
        chop_start = chop_end - MINBULK
        i = np.arange(len(phi))
        phi = phi[(i <= chop_start) | (i > chop_end)]

    if disp:
        print("SCFsolve execution time:", round(time() - starttime, 3), "s")

    return phi


_SZdist_dict = OrderedDict()


def SZdist(pdi, nn, cache=_SZdist_dict):
    """Calculate Shultz-Zimm distribution from PDI and number average DP

    Shultz-Zimm is a "realistic" distribution for linear polymers. Numerical
    problems arise when the distribution gets too uniform, so if we find them,
    default to an exact uniform calculation.
    """
    from scipy.special import gammaln

    args = pdi, nn
    if args in cache:
        cache[args] = cache.pop(args)
        return cache[args]

    uniform = False

    if pdi == 1.0:
        uniform = True
    elif pdi < 1.0:
        raise ValueError("Invalid PDI")
    else:
        x = 1.0 / (pdi - 1.0)
        # Calculate the distribution in chunks so we don't waste CPU time
        chunk = 256
        p_ni_list = []
        pdi_underflow = False

        for i in range(max(1, int((100 * nn) / chunk))):
            ni = np.arange(chunk * i + 1, chunk * (i + 1) + 1, dtype=np.float64)
            r = ni / nn
            xr = x * r

            p_ni = exp(log(x / ni) - gammaln(x + 1) + xr * (log(xr) / r - 1))

            pdi_underflow = (p_ni >= 1.0).any()  # catch "too small PDI"
            if pdi_underflow:
                break  # and break out to uniform calculation

            # Stop calculating when species account for less than 1ppm
            keep = (r < 1.0) | (p_ni >= 1e-6)
            if keep.all():
                p_ni_list.append(p_ni)
            else:
                p_ni_list.append(p_ni[keep])
                break
        else:  # Belongs to the for loop. Executes if no break statement runs.
            raise RuntimeError("SZdist overflow")

    if uniform or pdi_underflow:
        # NOTE: rounding here allows nn to be a double in the rest of the logic
        p_ni = np.zeros(int(round(nn)))
        p_ni[-1] = 1.0
    else:
        p_ni = np.concatenate(p_ni_list)
        p_ni /= p_ni.sum()
    cache[args] = p_ni

    if len(cache) > 9000:
        cache.popitem(last=False)

    return p_ni


def default_guess(segments=100, sigma=0.5, phi_b=0.1, chi=0, chi_s=0):
    """Produce an initial guess for phi via analytical approximants.

    For now, a line using numbers from scaling theory
    """
    ss = sqrt(sigma)
    default_layers = int(round(max(MINLAT, segments * ss)))
    default_phi0 = np.linspace(ss, phi_b, num=default_layers)
    return default_phi0


def SCFeqns(phi_z, chi, chi_s, sigma, n_avg, p_i, phi_b=0):
    """System of SCF equation for terminally attached polymers.

    Formatted for input to a nonlinear minimizer or solver.

    The sign convention here on u is "backwards" and always has been.
    It saves a few sign flips, and looks more like Cosgrove's.
    """

    # let the solver go negative if it wants
    phi_z = abs(phi_z)

    # penalize attempts that overfill the lattice
    toomuch = phi_z > 0.99999
    penalty_flag = toomuch.any()
    if penalty_flag:
        penalty = np.where(toomuch, phi_z - 0.99999, 0)
        phi_z[toomuch] = 0.99999

    # calculate new g_z (Boltzmann weighting factors)
    u_prime = log((1.0 - phi_z) / (1.0 - phi_b))
    u_int = 2 * chi * (old_correlate(phi_z, LAMBDA_ARRAY, 1) - phi_b)
    u_int[0] += chi_s
    u_z = u_prime + u_int
    g_z = exp(u_z)

    # normalize g_z for numerical stability
    u_z_avg = np.mean(u_z)
    g_z_norm = g_z / exp(u_z_avg)

    phi_z_new = calc_phi_z(g_z_norm, n_avg, sigma, phi_b, u_z_avg, p_i)

    eps_z = phi_z - phi_z_new

    if penalty_flag:
        np.copysign(penalty, eps_z, penalty)
        eps_z += penalty

    return eps_z


def calc_phi_z(g_z, n_avg, sigma, phi_b, u_z_avg=0, p_i=None):
    if p_i is None:
        segments = n_avg
        uniform = True
    else:
        segments = p_i.size
        uniform = segments == round(n_avg)

    g_zs = Propagator(g_z, segments)

    # for terminally attached chains
    if sigma:
        g_zs_ta = g_zs.ta()

        if uniform:
            c_i_ta = sigma / np.sum(g_zs_ta[:, -1])
            g_zs_ta_ngts = g_zs.ngts_u(c_i_ta)
        else:
            c_i_ta = sigma * p_i / np.sum(g_zs_ta, axis=0)
            g_zs_ta_ngts = g_zs.ngts(c_i_ta)

        phi_z_ta = compose(g_zs_ta, g_zs_ta_ngts, g_z)
    else:
        phi_z_ta = 0

    # for free chains
    if phi_b:
        g_zs_free = g_zs.free()

        if uniform:
            r_i = segments
            c_free = phi_b / r_i
            normalizer = exp(u_z_avg * r_i)
            c_free = c_free * normalizer
            g_zs_free_ngts = g_zs.ngts_u(c_free)
        else:
            r_i = np.arange(1, segments + 1)
            c_i_free = phi_b * p_i / r_i
            normalizer = exp(u_z_avg * r_i)
            c_i_free = c_i_free * normalizer
            g_zs_free_ngts = g_zs.ngts(c_i_free)

        phi_z_free = compose(g_zs_free, g_zs_free_ngts, g_z)
    else:
        phi_z_free = 0

    return phi_z_ta + phi_z_free


def compose(g_zs, g_zs_ngts, g_z):
    prod = g_zs * np.fliplr(g_zs_ngts)
    prod[np.isnan(prod)] = 0
    return np.sum(prod, axis=1) / g_z


class Propagator(object):
    def __init__(self, g_z, segments):
        self.g_z = g_z
        self.shape = int(g_z.size), int(segments)

    def ta(self):
        # terminally attached beginnings
        # forward propagator

        g_zs = self._new()
        g_zs[:, 0] = 0.0
        g_zs[0, 0] = self.g_z[0]
        _calc_g_zs_uniform(self.g_z, g_zs, LAMBDA_0, LAMBDA_1)
        return g_zs

    def free(self):
        # free beginnings
        # forward propagator

        g_zs = self._new()
        g_zs[:, 0] = self.g_z
        _calc_g_zs_uniform(self.g_z, g_zs, LAMBDA_0, LAMBDA_1)
        return g_zs

    def ngts_u(self, c):
        # free ends of uniform chains
        # reverse propagator

        g_zs = self._new()
        g_zs[:, 0] = c * self.g_z
        _calc_g_zs_uniform(self.g_z, g_zs, LAMBDA_0, LAMBDA_1)
        return g_zs

    def ngts(self, c_i):
        # free ends of disperse chains
        # reverse propagator

        g_zs = self._new()
        g_zs[:, 0] = c_i[-1] * self.g_z
        _calc_g_zs(self.g_z, c_i, g_zs, LAMBDA_0, LAMBDA_1)
        return g_zs

    def _new(self):
        return np.empty(self.shape, order="F")


try:
    from numba import njit

    USE_NUMBA = True
except ImportError:
    USE_NUMBA = False

# USE_NUMBA = False # Uncomment when doing timing tests

if USE_NUMBA:

    @njit("(f8[:], f8[:, :], f8, f8)", cache=True)
    def _calc_g_zs_uniform(g_z, g_zs, f0, f1):
        points, segments = g_zs.shape
        for r in range(segments - 1):
            g_zs[0, r + 1] = (g_zs[0, r] * f0 + g_zs[1, r] * f1) * g_z[0]
            # g_zs[1:-1, r+1] = np.correlate(g_zs[:, r], [f1, f0, f1], 'valid')*g_z[1:-1]
            for k in range(1, points - 1):
                g_zs[k, r + 1] = (g_zs[k, r] * f0 + (g_zs[k - 1, r] + g_zs[k + 1, r]) * f1) * g_z[k]
            g_zs[-1, r + 1] = (g_zs[-2, r] * f1 + g_zs[-1, r] * f0) * g_z[-1]

    @njit("(f8[:], f8[:], f8[:, :], f8, f8)", cache=True)
    def _calc_g_zs(g_z, c_i, g_zs, f0, f1):
        points, segments = g_zs.shape
        for r in range(segments - 1):
            c_ir = c_i[segments - (r + 1) - 1]
            g_zs[0, r + 1] = (g_zs[0, r] * f0 + g_zs[1, r] * f1 + c_ir) * g_z[0]
            # g_zs[1:-1, r+1] = (np.correlate(g_zs[:, r], fir, 'valid') + c_ir)*g_z[1:-1]
            for k in range(1, points - 1):
                g_zs[k, r + 1] = (g_zs[k, r] * f0 + (g_zs[k - 1, r] + g_zs[k + 1, r]) * f1 + c_ir) * g_z[k]
            g_zs[-1, r + 1] = (g_zs[-2, r] * f1 + g_zs[-1, r] * f0 + c_ir) * g_z[-1]

else:

    def _calc_g_zs(g_z, c_i, g_zs, f0, f1):
        coeff = np.array([f1, f0, f1])
        pg_zs = g_zs[:, 0]
        segment_iterator = enumerate(c_i[::-1])
        next(segment_iterator)
        for r, c in segment_iterator:
            g_zs[:, r] = pg_zs = (old_correlate(pg_zs, coeff, 1) + c) * g_z

    def _calc_g_zs_uniform(g_z, g_zs, f0, f1):
        coeff = np.array([f1, f0, f1])
        segments = g_zs.shape[1]
        pg_zs = g_zs[:, 0]
        for r in range(1, segments):
            g_zs[:, r] = pg_zs = old_correlate(pg_zs, coeff, 1) * g_z



import asyncio
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, Union

# import bumps.webview.server.api as bumps_api
import numpy as np
from bumps.errplot import error_points_from_state
from bumps.webview.server.api import (
    add_notification,
    get_chisq,
    log,
    logger,
    now_string,
    register,
    state,
    to_json_compatible_dict,
)

from refl1d.uncertainty import calc_errors
from refl1d.experiment import Experiment, ExperimentBase, MixedExperiment
from refl1d.probe.data_loaders.load4 import load4
from refl1d.probe import PolarizedNeutronProbe
from .profile_uncertainty import show_errors
from .profile_plot import plot_multiple_sld_profiles, ModelSpec

# state.problem.serializer = "dataclass"


@register
async def get_plot_data(view: str = "linear"):
    # TODO: implement view-dependent return instead of doing this in JS
    # (calculate x,y,dy.dx for given view, excluding log)
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem
    chisq = await get_chisq(fitProblem)
    plotdata = []
    result = {"chisq": chisq, "plotdata": plotdata}
    for model in fitProblem.models:
        assert isinstance(model, ExperimentBase)
        theory = model.reflectivity()
        probe = model.probe
        probe_data = get_probe_data(theory, probe, model._substrate, model._surface)
        plotdata.append(probe_data)

    return to_json_compatible_dict(result)


async def create_profile_plots(model_specs: List[ModelSpec]):
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem
    plot_items = []
    color_index = 0
    for model_index, model in enumerate(fitProblem.models):
        for sample_index, part in enumerate(getattr(model, "parts", [model])):
            spec = dict(model_index=model_index, sample_index=sample_index)
            if spec in model_specs:
                plot_item = dict(model=part, spec=spec, color_index=color_index)
                plot_items.append(plot_item)
            color_index += 1

    fig = plot_multiple_sld_profiles(plot_items)
    return fig


@register
async def get_profile_plots(model_specs: List[ModelSpec]):
    fig = await create_profile_plots(model_specs)
    output = to_json_compatible_dict(fig.to_dict())
    del fig
    return output


def get_single_probe_data(theory, probe, substrate=None, surface=None, polarization=""):
    fresnel_calculator = probe.fresnel(substrate, surface)
    Q, FQ = probe.apply_beam(probe.calc_Q, fresnel_calculator(probe.calc_Qo))
    Q, R = theory
    output: Dict[str, Union[str, np.ndarray]]
    assert isinstance(FQ, np.ndarray)
    if len(Q) != len(probe.Q):
        # Saving interpolated data
        output = dict(Q=Q, theory=R, fresnel=np.interp(Q, probe.Q, FQ))
    elif getattr(probe, "R", None) is not None:
        output = dict(
            Q=probe.Q,
            dQ=probe.dQ,
            R=probe.R,
            dR=probe.dR,
            theory=R,
            fresnel=FQ,
            background=probe.background.value,
            intensity=probe.intensity.value,
        )
    else:
        output = dict(Q=probe.Q, dQ=probe.dQ, theory=R, fresnel=FQ)
    output["polarization"] = polarization
    output["label"] = probe.label()
    return output


def get_probe_data(theory, probe, substrate=None, surface=None):
    if isinstance(probe, PolarizedNeutronProbe):
        output = []
        for xsi, xsi_th, suffix in zip(probe.xs, theory, ("--", "-+", "+-", "++")):
            if xsi is not None:
                output.append(get_single_probe_data(xsi_th, xsi, substrate, surface, suffix))
        return output
    else:
        return [get_single_probe_data(theory, probe, substrate, surface)]


@register
async def get_model_names():
    problem = state.problem.fitProblem
    if problem is None:
        return None
    output: List[Dict] = []
    for model_index, model in enumerate(problem.models):
        if isinstance(model, Experiment):
            output.append(dict(name=model.name, part_name=None, model_index=model_index, part_index=0))
        elif isinstance(model, MixedExperiment):
            for part_index, part in enumerate(model.parts):
                output.append(
                    dict(name=model.name, part_name=part.name, model_index=model_index, part_index=part_index)
                )
    return output


@lru_cache(maxsize=30)
def _get_profile_uncertainty_plot(
    auto_align: bool = True,
    align: float = 0.0,
    nshown: int = 5000,
    npoints: int = 5000,
    random: bool = True,
    residuals: bool = False,
):
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = deepcopy(state.problem.fitProblem)
    uncertainty_state = state.fitting.uncertainty_state
    align_arg = "auto" if auto_align else align
    if uncertainty_state is not None:
        import time

        start_time = time.time()
        logger.info(f"queueing new profile uncertainty plot... {start_time}")
        error_points = error_points_from_state(uncertainty_state, nshown=nshown, random=random, portion=1.0)
        logger.info(f"points calculated: {time.time() - start_time}")
        errs = calc_errors(fitProblem, error_points)
        logger.info(f"errors calculated: {time.time() - start_time}")
        error_result = show_errors(errs, npoints=npoints, align=align_arg, residuals=residuals)
        error_result["fig"] = error_result["fig"].to_dict()
        logger.info(f"time to render but not serialize... {time.time() - start_time}")
        output = to_json_compatible_dict(error_result)
        del error_result
        end_time = time.time()
        logger.info(f"time to draw profile uncertainty plot: {end_time - start_time}")
        return output
    else:
        return None


@register
async def get_profile_uncertainty_plot(
    auto_align: bool = True,
    align: float = 0.0,
    nshown: int = 5000,
    npoints: int = 5000,
    random: bool = True,
    residuals: bool = False,
):
    result = await asyncio.to_thread(
        _get_profile_uncertainty_plot,
        auto_align=auto_align,
        align=align,
        nshown=nshown,
        npoints=npoints,
        random=random,
        residuals=residuals,
    )
    return result


@register
async def load_probe_from_file(pathlist: List[str], filename: str, model_index: int = 0, fwhm: bool = True):
    path = Path(*pathlist)
    fitProblem = state.problem.fitProblem if state.problem is not None else None
    if fitProblem is None:
        await log("Error: Can't load data if no problem defined")
    else:
        models = list(fitProblem.models)
        num_models = len(models)
        if model_index >= num_models:
            await log(f"Error: Can not access model at model_index {model_index} (only {num_models} defined)")
            return
        model: Experiment = models[model_index]
        probe = load4(str(path / filename), FWHM=fwhm)
        model.probe = probe
        fitProblem.model_reset()
        fitProblem.model_update()
        state.save()
        state.shared.updated_model = now_string()
        state.shared.updated_parameters = now_string()
        await add_notification(content=f"from {filename} to model {model_index}", title="Data loaded:", timeout=2000)



"""
A base panel to draw the profile
"""

import os

import numpy as np
import wx
from bumps.fitproblem import FitProblem
from bumps.gui import signal
from matplotlib.axes import Subplot
from matplotlib.backend_bases import FigureManagerBase
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg
from matplotlib.figure import Figure

from refl1d.experiment import MixedExperiment

# from .binder import pixel_to_data
from .interactor import BaseInteractor
from .profilei import ProfileInteractor
from .util import CopyImage

IS_MAC = wx.Platform == "__WXMAC__"


# ------------------------------------------------------------------------
class ModelView(wx.Panel):
    title = "Profile"
    default_size = (600, 400)

    def __init__(self, *args, **kw):
        wx.Panel.__init__(self, *args, **kw)

        # Fig
        self.fig = Figure(
            figsize=(1, 1),
            dpi=75,
            facecolor="white",
            edgecolor="white",
        )
        # Canvas
        self.canvas = FigureCanvas(self, -1, self.fig)
        self.fig.set_canvas(self.canvas)

        # Axes
        self.axes = self.fig.add_axes(Subplot(self.fig, 111))
        self.axes.set_autoscale_on(False)
        self.theta_axes = self.axes.twinx()
        self.theta_axes.set_autoscale_on(False)

        # Show toolbar or not?
        self.toolbar = NavigationToolbar2WxAgg(self.canvas)
        self.toolbar.Show(True)

        # Create a figure manager to manage things
        self.figmgr = FigureManagerBase(self.canvas, 1)

        # Panel layout
        self.profile_selector_label = wx.StaticText(self, label="Sample")
        self.profile_selector = wx.Choice(self)
        self.profile_selector.Hide()
        self.profile_selector_label.Hide()
        self.Bind(wx.EVT_CHOICE, self.OnProfileSelect)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.canvas, 1, border=2, flag=wx.LEFT | wx.TOP | wx.GROW)
        self.tbsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.tbsizer.Add(self.toolbar, 0, wx.ALIGN_CENTER_VERTICAL)
        self.tbsizer.AddSpacer(20)
        self.tbsizer.Add(self.profile_selector_label, 0, wx.ALIGN_CENTER_VERTICAL)
        self.tbsizer.AddSpacer(5)
        self.tbsizer.Add(self.profile_selector, 0, wx.ALIGN_CENTER_VERTICAL)
        self.sizer.Add(self.tbsizer)

        self.SetSizer(self.sizer)
        self.sizer.Fit(self)

        # Status bar
        frame = self.GetTopLevelParent()
        self.statusbar = frame.GetStatusBar()
        if self.statusbar is None:
            self.statusbar = frame.CreateStatusBar()

        def status_update(msg):
            return self.statusbar.SetStatusText(msg)

        # Set the profile interactor
        self.profile = ProfileInteractor(self.axes, self.theta_axes, status_update=status_update)

        # Add context menu and keyboard support to canvas
        self.canvas.Bind(wx.EVT_RIGHT_DOWN, self.OnContextMenu)
        # self.canvas.Bind(wx.EVT_LEFT_DOWN, lambda evt: self.canvas.SetFocus())

        self.model = None
        self._need_interactors = self._need_redraw = False
        self.Bind(wx.EVT_SHOW, self.OnShow)

    def OnContextMenu(self, event):
        """
        Forward the context menu invocation to profile, if profile exists.
        """
        sx, sy = event.GetX(), event.GetY()
        # transform = self.axes.transData
        # data_x,data_y = pixel_to_data(transform, sx, self.fig.bbox.height-sy)

        popup = wx.Menu()
        item = popup.Append(wx.ID_ANY, "&Grid on/off", "Toggle grid lines")
        wx.EVT_MENU(self, item.GetId(), lambda _: (self.axes.grid(), self.fig.canvas.draw_idle()))
        item = popup.Append(wx.ID_ANY, "&Rescale", "Show entire profile")
        wx.EVT_MENU(self, item.GetId(), lambda _: (self.profile.reset_limits(), self.profile.draw_idle()))
        self.PopupMenu(popup, (sx, sy))
        return False

    def OnProfileSelect(self, event):
        self._set_profile(*self.profiles[event.GetInt()])

    # ==== Model view interface ===
    def OnShow(self, event):
        if not event.Show:
            return
        # print "showing profile"
        if self._need_redraw:
            self.redraw(reset_interactors=False, reset_limits=True)
        # event.Skip()

    def get_state(self):
        return self.model

    def set_state(self, state):
        self.set_model(state)

    def set_model(self, model):
        # print ">>>>>>> refl1d profile set model"
        self.model = model
        self.redraw(reset_interactors=True, reset_limits=True)

    def update_model(self, model):
        # print ">>>>>>> refl1d profile update model"
        if self.model == model:
            self.redraw(reset_interactors=False, reset_limits=True)

    def update_parameters(self, model):
        # print ">>>>>>> refl1d profile update parameters"
        if self.model == model:
            self.redraw(reset_interactors=False, reset_limits=False)

    def redraw(self, reset_interactors=False, reset_limits=False):
        if reset_interactors:
            self._need_interactors = True

        if not self.IsShown():
            self._need_redraw = True
            return

        if self._need_interactors:
            self._create_interactors()
            self._set_profile(*self.profiles[0])
            self._need_interactors = False

        self.profile.redraw(reset_limits=reset_limits)

    # =============================================
    def _create_interactors(self):
        self.profiles = []

        def add_profiles(name, exp, idx):
            if isinstance(exp, MixedExperiment):
                for i, p in enumerate(exp.parts):
                    self.profiles.append((name + chr(ord("a") + i), p, idx))
            else:
                self.profiles.append((name, exp, idx))

        if isinstance(self.model, FitProblem):
            for i, p in enumerate(self.model.models):
                if hasattr(p, "reflectivity"):
                    name = p.name
                    if not name:
                        name = "M%d" % (i + 1)
                    add_profiles(name, p, i)
        else:
            add_profiles("", self.model, -1)

        self.profile_selector.Clear()
        if len(self.profiles) > 1:
            self.profile_selector.AppendItems([k for k, _, _ in self.profiles])
            self.profile_selector_label.Show()
            self.profile_selector.Show()
            self.profile_selector.SetSelection(0)
        else:
            self.profile_selector_label.Hide()
            self.profile_selector.Hide()

    def _set_profile(self, name, experiment, idx):
        # Turn the model into a user interface
        # It is the responsibility of the party that is indicating
        # that a redraw is necessary to clear the precalculated
        # parts of the view; otherwise the theory function calculator
        # is going to be triggered twice.  This happens inside profile
        # before the profile is calculated.  Note that the profile
        # panel will receive its own signal, which will cause the
        # profile interactor to draw itself again.  We hope this isn't
        # too much of a problem.
        def signal_update():
            """Notify other views that the model has changed"""
            signal.update_parameters(model=self.model)

        def force_recalc():
            self.model.model_update()

        if isinstance(self.model, FitProblem):
            self.model.set_active_model(idx)
        self.profile.set_experiment(experiment, force_recalc=force_recalc, signal_update=signal_update)

    def onPrinterSetup(self, event=None):
        self.canvas.Printer_Setup(event=event)

    def onPrinterPreview(self, event=None):
        self.canvas.Printer_Preview(event=event)

    def onPrint(self, event=None):
        self.canvas.Printer_Print(event=event)

    def OnSaveFigureMenu(self, evt):
        """
        Save the current figure as an image file
        """
        dlg = wx.FileDialog(
            self,
            message="Save Figure As ...",
            defaultDir=os.getcwd(),
            defaultFile="",
            wildcard="PNG files (*.png)|*.png|BMP files (*.bmp)|*.bmp|All files (*.*)|*.*",
            style=wx.SAVE,
        )
        _val = dlg.ShowModal()
        if _val == wx.ID_CANCEL:
            return  # Do nothing
        if _val == wx.ID_OK:
            outfile = dlg.GetPath()

        dlg.Destroy()

        # Save
        self.fig.savefig(outfile)

    def GetToolBar(self):
        """
        backend_wx call this function. KEEP it
        """
        return None

    def GetTitle(self):
        """
        backend_wx calls this function.
        """
        return self.title

    def OnPanelFrameClose(self, evt):
        """
        On Close this Frame
        """
        self.Destroy()
        evt.Skip()

    def OnCopyFigureMenu(self, evt):
        """
        Copy the current figure
        """
        CopyImage(self.canvas)

    def CanShowContextMenu(self):
        return True

    def quit_on_error(self):
        np.seterr(all="raise")
        ProfileInteractor._debug = True
        BaseInteractor._debug = True



# This program is in the public domain
# Author: Paul Kienzle
"""
NCNR data loaders

The following instruments are defined:

    MAGIK, PBR, ANDR, NG1, NG7 and XRay

These are :class:`refl1d.probe.instrument.Monochromatic` classes tuned with default
instrument parameters and loaders for reduced NCNR data.

The instruments can be used to load data or to compute resolution functions
for the purposes.

Example loading data:

    >>> import numpy as np
    >>> import pylab
    >>> from refl1d.names import Experiment, NCNR, air, gold, permalloy, sample_data, silicon
    >>> datafile = sample_data('chale207.refl')
    >>> instrument = NCNR.ANDR(Tlo=0.5, slits_at_Tlo=0.2, slits_below=0.1)
    >>> probe = instrument.load(datafile)
    >>> probe.plot(view='log')

Magnetic data has multiple cross sections and often has fixed slits:

    >>> datafile = sample_data('lha03_255G.refl')
    >>> instrument = NCNR.NG1(slits_at_Tlo=1)
    >>> probe = instrument.load_magnetic(datafile)
    >>> probe.plot(view='SA', substrate=silicon) # Spin asymmetry view

For simulation, you need a probe and a sample:

    >>> instrument = NCNR.ANDR(Tlo=0.5, slits_at_Tlo=0.2, slits_below=0.1)
    >>> probe = instrument.probe(T=np.linspace(0, 5, 51))
    >>> probe.plot_resolution()
    >>> sample = silicon(0, 10) | gold(100, 10) | air
    >>> M = Experiment(probe=probe, sample=sample)
    >>> M.simulate_data() # Optional
    >>> M.plot()

And for magnetic:

    >>> instrument = NCNR.NG1(slits_at_Tlo=1)
    >>> #sample = silicon(0, 10) | Magnetic(permalloy(100, 10), rho_M=3) | air
    >>> #M = Experiment(probe=probe, sample=sample)
    >>> #M.simulate_data()
    >>> #M.plot()
    >>> #probe = instrument.simulate_magnetic(sample, T=np.linspace(0, 5, 51))
    >>> #h = pylab.plot(probe.Q, probe.dQ)
    >>> #h = pylab.ylabel('resolution (1-sigma)')
    >>> #h = pylab.xlabel('Q (inv A)')

See :mod:`instrument <refl1d.probe.instrument>` for details.
"""

import os

from bumps.data import parse_file

from ...sample.reflectivity import BASE_GUIDE_ANGLE
from ..instrument import Monochromatic
from ..probe import PolarizedNeutronProbe


def load(filename, instrument=None, **kw):
    """
    Return a probe for NCNR data.

    Keyword arguments are as specified Monochromatic instruments.
    """
    if filename is None:
        return None
    if instrument is None:
        instrument = Monochromatic()
    header, data = parse_ncnr_file(filename)
    # calling parameters override what's in the file.
    header.update(filename=filename, **kw)
    Q, R, dR = data
    header.pop("Q", None)  # if columns are preceded by # Q R dR
    probe = instrument.probe(Q=Q, data=(R, dR), **header)
    probe.title = header["title"] if "title" in header else filename
    probe.date = header["date"] if "date" in header else "unknown"
    probe.instrument = header["instrument"] if "instrument" in header else instrument.instrument
    return probe


def load_magnetic(filename, Aguide=BASE_GUIDE_ANGLE, H=0, shared_beam=True, **kw):
    """
    Return a probe for magnetic NCNR data.

    *filename* (string, or 4x string)
        If it is a string, then filenameA, filenameB, filenameC, filenameD,
        are the --, -+, +-, ++ cross sections, otherwise the individual
        cross sections should the be the file name for the cross section or
        None if the cross section does not exist.
    *Aguide* (degrees)
        Angle of the guide field relative to the beam.  270 is the default.
    *shared_beam* (True)
        Use false if beam parameters should be fit separately for the
        individual cross sections.

    Other keyword arguments are for the individual cross section loaders
    as specified in
    :class:`instrument.Monochromatic <refl1d.probe.instrument.Monochromatic>`.

    The data sets should are the base filename with an additional character
    corresponding to the spin state::

        'a' corresponds to spin --
        'b' corresponds to spin -+
        'c' corresponds to spin +-
        'd' corresponds to spin ++

    Unfortunately the interpretation is a little more complicated than
    this as the data acquisition system assigns letter on the basis of
    flipper state rather than neutron spin state.  Whether flipper on
    or off corresponds to spin up or down depends on whether the
    polarizer/analyzer is a supermirror in transmission or reflection
    mode, or in the case of ^3He polarizers, whether the polarization
    is up or down.

    For full control, specify filename as a list of files, with None
    for the missing cross sections.
    """
    probes = [load(v, **kw) for v in find_xsec(filename)]
    if all(p is None for p in probes):
        raise IOError("Data set has no magnetic cross sections: %r" % filename)
    probe = PolarizedNeutronProbe(probes, Aguide=Aguide, H=H)
    if shared_beam:
        probe.shared_beam()  # Share the beam parameters by default
    return probe


def find_xsec(filename):
    """
    Find files containing the polarization cross-sections.

    Returns tuple with file names for ++ +- -+ -- cross sections, or
    None if the spin cross section does not exist.
    """
    # Check if it is a string.  If not, assume it is a length 4 tuple
    try:
        filename + "s"
    except Exception:
        return filename

    if filename[-1] in "abcdABCD":
        filename = filename[:-1]

    def check(a):
        if os.path.exists(filename + a):
            return filename + a
        elif os.path.exists(filename + a.lower()):
            return filename + a.lower()
        else:
            return None

    return check("A"), check("B"), check("C"), check("D")


def parse_ncnr_file(filename):
    """
    Parse NCNR reduced data file returning *header* and *data*.

    *header* dictionary of fields such as 'data', 'title', 'instrument'
    *data* 2D array of data

    If 'columns' is present in header, it will be a list of the names of
    the columns.  If 'instrument' is present in the header, the default
    instrument geometry will be specified.

    Slit geometry is set to the default from the instrument if it is not
    available in the reduced file.
    """
    header, data = parse_file(filename)

    # Fill in instrument parameters, if not available from the file
    if "instrument" in header and header["instrument"] in INSTRUMENTS:
        instrument = INSTRUMENTS[header["instrument"]]
        header.setdefault("radiation", instrument.radiation)
        header.setdefault("wavelength", str(instrument.wavelength))
        header.setdefault("dLoL", str(instrument.dLoL))
        header.setdefault("d_s1", str(instrument.d_s1))
        header.setdefault("d_s2", str(instrument.d_s2))

    if "columns" in header:
        header["columns"] = header["columns"].split()
    for key in ("wavelength", "dLoL", "d_s1", "d_s2"):
        if key in header:
            header[key] = float(header[key])

    return header, data


class NCNRData(object):
    def readfile(self, filename):
        return parse_ncnr_file(filename)

    def load(self, filename, **kw):
        return load(filename, instrument=self, **kw)

    def load_magnetic(self, filename, **kw):
        return load_magnetic(filename, instrument=self, **kw)


class MAGIK(NCNRData, Monochromatic):
    """
    Instrument definition for NCNR MAGIK diffractometer/reflectometer.
    """

    instrument = "MAGIK"
    radiation = "neutron"
    wavelength = 5.0042
    dLoL = 0.009
    d_s1 = 1321.0 + 438.0
    d_s2 = 1321.0 - 991.0


class PBR(NCNRData, Monochromatic):
    """
    Instrument definition for NCNR PBR reflectometer.
    """

    instrument = "PBR"
    radiation = "neutron"
    wavelength = 4.75
    dLoL = 0.015
    d_s1 = 1835
    d_s2 = 343
    d_s3 = 380
    d_s4 = 1015


class NG7(NCNRData, Monochromatic):
    """
    Instrument definition for NCNR NG-7 reflectometer.
    """

    instrument = "NG-7"
    radiation = "neutron"
    wavelength = 4.768
    dLoL = 0.025  # 2.5% FWHM wavelength spread
    d_s2 = 222.25
    d_s1 = 1722.25
    d_detector = 2000.0


class XRay(NCNRData, Monochromatic):
    """
    Instrument definition for NCNR X-ray reflectometer.

    Normal dT is in the range 2e-5 to 3e-4.

    Slits are fixed throughout the experiment in one of a
    few preconfigured openings.  Please update this file with
    the standard configurations when you find them.

    You can choose to ignore the geometric calculation entirely
    by setting the slit opening to 0 and using sample_broadening
    to define the entire divergence.  Note that Probe.sample_broadening
    is a fittable parameter, so you need to access its value::

        >>> from refl1d.names import sample_data, NCNR
        >>> file = sample_data("spin_valve01.refl")
        >>> xray = NCNR.XRay(slits_at_Tlo=0)
        >>> data = xray.load(file, sample_broadening=1e-4)
        >>> print(data.sample_broadening.value)
        0.0001
    """

    instrument = "X-ray"
    radiation = "xray"
    wavelength = 1.5416
    dLoL = 1e-3 / wavelength
    d_s1 = 275.5
    d_s2 = 192.5
    d_s3 = 175.0
    d_detector = None


class ANDR(NCNRData, Monochromatic):
    """
    Instrument definition for NCNR AND/R diffractometer/reflectometer.
    """

    instrument = "AND/R"
    radiation = "neutron"
    wavelength = 5.0042
    dLoL = 0.009
    d_s1 = 230.0 + 1856.0
    d_s2 = 230.0


class NG1(NCNRData, Monochromatic):
    """
    Instrument definition for NCNR NG-1 reflectometer.
    """

    instrument = "NG-1"
    radiation = "neutron"
    wavelength = 4.75
    dLoL = 0.015
    d_s1 = 75 * 25.4
    d_s2 = 14 * 25.4
    d_s3 = 9 * 25.4
    d_s4 = 42 * 25.4


# Instrument names assigned by reflpak
INSTRUMENTS = {
    "CG-D": MAGIK,
    "NG-D": PBR,
    "CG-1": ANDR,
    "NG-1": NG1,
    "NG-7": NG7,
    "Xray": XRay,
}

_ = r'''
def _counting_time(instrument, sample, uncertainty,
    Qrange, Qstep, beam_rate, num_parts):
    r"""
    Simulate counting time for a particular sample.

    :Parameters:
        *sample* : Stack
            Model of the sample.
        *uncertainty* = 0.01 : float
            Relative uncertainty in the measurement.

    Additional :meth:`probe` keyword parameters are required to define
    the set of angles to be measured

    Returns
    -------

    *experiment* : Experiment
        Sample + probe with simulated data.

    Algorithm
    ---------

    Assuming our counts follow approximately the Fresnel reflectivity
    of the sample, $F$, and we are targeting an fractional uncertainty
    $\Delta R/R = \sigma$, we can calculate the desired incident beam
    $I = 1/(F\sigma^2)$ that will yield this uncertainty.  With $I$,
    we can compute the  expected number of counts on the detector due
    to reflection off the sample (this is just $R_{\rm th} I$) and use
    that to simulate detector counts $D$ by drawing from a Poisson
    distribution $D ~ P(R_{\rm th} I)$.  Given $D$ and $I$ we can use
    the normal reflectometry reduction process to get $(R, \Delta R)$
    as:

    .. math:

        I &=& 1/(F \sigma^2) //
        D &~& P(R_{\rm th} I) //
        R &=& D/I //
        \Delta R &=& \sqrt(D)/I
    """
    import numpy as np
    from refl1d.names.experiment import Experiment
    probe = self.probe(**kw)
    M = Experiment(probe=probe, sample=sample)
    if 1: # Fresnel counting
        I = 1/(M.fresnel()* uncertainty**2)
    else: # Q^4 counting
        I = 1/(100*probe.Q**4 * uncertainty**2)
    D = np.random.poisson( Rth * I )
    R, dR = D/I, np.sqrt(D)/I
    probe.data = R, dR

    return M
'''



def clip(a, lo, hi):
    return max(lo, min(a, hi))


def setpar(p, v):
    p.value = clip(v, *p.prior.limits)


def CopyImage(canvas):
    """
    Copy the MPL canvas to the wx clipboard.
    """
    import wx

    bmp = wx.BitmapDataObject()
    bmp.SetBitmap(canvas.bitmap)

    wx.TheClipboard.Open()
    wx.TheClipboard.SetData(bmp)
    wx.TheClipboard.Close()


def showErrorMsg(parent, msg, title):
    """
    Show error message
    """
    import wx

    msg = wx.MessageDialog(parent, msg, title, wx.ICON_ERROR | wx.OK)
    msg.ShowModal()
    msg.Destroy()


def showWarningMsg(parent, msg, title):
    """
    Show Warning message
    """
    import wx

    msg = wx.MessageDialog(parent, msg, title, wx.ICON_WARNING | wx.OK)
    msg.ShowModal()
    msg.Destroy()


def CheckValid(s):
    """
    Translate the string into float, and check it.
    """
    try:
        val = float(s)
    except:
        val = 1.0
        msg = "%s can't change to float" % (s)
        raise ValueError(msg)

    return val


def twinx(ax=None):
    """
    Make a second axes overlay ax
    (or the current axes if ax is None) sharing the xaxis.

    The ticks for ax2 will be placed on the right,
    and the ax2 instance is returned.  See examples/two_scales.py

    Warning: This is a function to simulate the pylab.twinx in WX
    """
    if ax is None:
        return None

    ax2 = ax.figure.add_axes(ax.get_position(), sharex=ax, frameon=False)
    ax2.yaxis.tick_right()
    ax2.yaxis.set_label_position("right")
    ax.yaxis.tick_left()

    return ax2


def clear_axes(ax):
    """
    Clear axes from a figure
    """
    ax.legend_ = None
    ax.set_title(r"")
    ax.set_xlabel(r"")
    ax.set_ylabel(r"")
    ax.figure.canvas.draw_idle()


def filterP(f):
    """
    Translate a obj parameter into "string" format
    """
    if hasattr(f, "build"):
        return f.build()
    else:
        return f


def decodeP(p):
    """
    Translate a parameter into obj format

    Decompose a parameter.
    In this code. It only works for spline obj
    """
    try:
        ret = eval(p)
    except:
        ret = p
    return ret



"""
interface  interactor.
"""

from .config import interface_color, pick_radius
from .interactor import BaseInteractor
from .util import clip, setpar

MAX_ROUGH = 1


class InterfaceInteractor(BaseInteractor):
    """
    Control the roughness of the layers.
    """

    def __init__(self, profile):
        BaseInteractor.__init__(self, profile)
        ax = profile.axes

        # markers for roughness
        style = dict(
            linestyle="",
            transform=profile.xcoords,
            marker="s",  # square
            markersize=7,
            color=interface_color,
            alpha=0.6,
            pickradius=pick_radius,
            zorder=8,  # Prefer this to other lines
            visible=False,
        )
        self.markers = [
            ax.plot([0], [0.05], label=label, **style)[0] for label in ("interface marker L", "interface marker R")
        ]

        # lines for roughness
        style = dict(linestyle="-", transform=profile.xcoords, marker="", color=interface_color, visible=False)
        self.lines = [
            ax.plot([0, 0], [0.05, 0.05], label=label, **style)[0] for label in ("interface line L", "interface line R")
        ]

        self.connect_markers(self.markers)
        self._left = self._right = None

    def set_layer(self):
        """
        Move markers to the new layer
        """
        n = self.profile.layer_num
        z = self.profile.boundary[1:-1]
        show_left = n is not None and n > 0
        show_right = n is not None and n < len(z)
        if show_left:
            self._left = self.profile.sample_layer(n - 1).interface
        else:
            self._left = None
        if show_right:
            self._right = self.profile.sample_layer(n).interface
        else:
            self._right = None
        # self.update_markers()

    def update_markers(self):
        """
        Draw the new roughness on the graph.
        """
        n = self.profile.layer_num
        z = self.profile.boundary[1:-1]
        show_left = self._left is not None
        show_right = self._right is not None

        self.markers[0].set(visible=show_left)
        self.lines[0].set(visible=show_left)
        if show_left:
            self.markers[0].set_xdata([z[n - 1] + self._left.value])
            self.lines[0].set_xdata([z[n - 1], z[n - 1] + self._left.value])

        self.markers[1].set(visible=show_right)
        self.lines[1].set(visible=show_right)
        if show_right:
            self.markers[1].set_xdata([z[n] - self._right.value])
            self.lines[1].set_xdata([z[n], z[n] - self._right.value])

    def clear_markers(self):
        """
        Remove interface markers from the graph.
        """
        BaseInteractor.clear_markers(self)
        for h in self.lines:
            h.remove()
        self.lines = []

    def save(self, event):
        """
        Remember the interface for this layer and the next so that we
        can restore on Esc.
        """

        if self._left is not None:
            self._left_value = self._left.value
        if self._right is not None:
            self._right_value = self._right.value

    def restore(self, event):
        """
        Restore the roughness for this layer.
        """
        if self._left is not None:
            self._left.value = self._left_value
        if self._right is not None:
            self._right.value = self._right_value

    def drag(self, event):
        """
        Process move to a new position, making sure that the move is allowed.
        """
        z = self.profile.boundary
        n = self.profile.layer_num
        side = self.markers.index(event.artist)
        if side == 0:  # Left
            limit = min(z[n] - z[n - 1], z[n + 1] - z[n])
            v = clip(event.xdata - z[n], 0, limit / MAX_ROUGH)
            setpar(self._left, v)

        else:  # Right
            limit = min(z[n + 1] - z[n], z[n + 2] - z[n + 1])
            v = clip(z[n + 1] - event.xdata, 0, limit / MAX_ROUGH)
            setpar(self._right, v)

        # self.update_markers()



"""
Exported names

In model definition scripts, rather than importing symbols one by one, you
can simply perform:

    from refl1d.names import *

This is bad style for library and applications but convenient for
small scripts.
"""

import logging
import sys

import numpy as np
from bumps import pmath
from bumps.fitproblem import FitProblem, MultiFitProblem
from bumps.parameter import FreeVariables, Parameter
from bumps.pdfwrapper import PDF
from periodictable import elements, formula

from .deprecated.magnetic import FreeMagnetic, MagneticSlab, MagneticStack, MagneticTwist
from .experiment import Experiment, MixedExperiment, plot_sample
from .probe.data_loaders import ncnrdata as NCNR
from .probe.data_loaders import snsdata as SNS
from .probe.data_loaders.load4 import load4
from .probe.data_loaders.stajconvert import load_mlayer, save_mlayer
from .probe.instrument import Monochromatic, Pulsed
from .probe import (
    NeutronProbe,
    PolarizedNeutronProbe,
    PolarizedQProbe,
    Probe,
    ProbeSet,
    QProbe,
    XrayProbe,
)
from .sample.cheby import ChebyVF, FreeformCheby, cheby_approx, cheby_points
from .sample.flayer import FunctionalMagnetism, FunctionalProfile
from .sample.layers import Slab, Stack
from .sample.magnetism import FreeMagnetism, FreeMagnetismInterface, Magnetism, MagnetismStack, MagnetismTwist
from .sample.material import SLD, Compound, Material, Mixture

# Pull in common materials for reflectometry experiments.
# This could lead to a lot of namespace pollution, and particularly to
# confusion if the user also does "from periodictable import *" since
# both of them create elements.
from .sample.materialdb import *
from .sample.mono import FreeInterface, FreeLayer
from .sample.polymer import EndTetheredPolymer, PolymerBrush, PolymerMushroom, VolumeProfile, layer_thickness
from .utils.support import sample_data


# Deprecated names
def ModelFunction(*args, **kw):
    raise NotImplementedError("ModelFunction no longer supported --- use PDF instead")


PolarizedNeutronQProbe = PolarizedQProbe
numpy = np



from math import erf, sqrt, exp

PI4 = 12.56637061435917295385
PI_180 = 0.01745329251994329576
LN256 = 5.54517744447956247533
SQRT2 = 1.41421356237309504880
SQRT2PI = 2.50662827463100050241
LOG_RESLIMIT = -6.90775527898213703123
root_12_over_2 = sqrt(3)

prange = range


def convolve_uniform(xi, yi, x, dx, y):
    left_index = 0
    N_xi = len(xi)
    N_x = len(x)
    for k in prange(N_x):
        x_k = x[k]
        # Convert 1-sigma width to 1/2 width of the region
        limit = dx[k] * root_12_over_2
        # print(f"point {x_k} +/- {limit}")
        # Find integration limits, bound by the range of the data
        left, right = max(x_k - limit, xi[0]), min(x_k + limit, xi[-1])
        if right < left:
            # Convolution does not overlap data range.
            y[k] = 0.0
            continue

        # Find the starting point for the convolution by first scanning
        # forward until we reach the next point greater than the limit
        # (we might already be there if the next output point has wider
        # resolution than the current point), then scanning backwards to
        # get to the last point before the limit. Make sure we have at
        # least one interval so that we don't have to check edge cases
        # later.
        while left_index < N_xi - 2 and xi[left_index] < left:
            left_index += 1
        while left_index > 0 and xi[left_index] > left:
            left_index -= 1

        # Set the first interval.
        total = 0.0
        right_index = left_index + 1
        x1, y1 = xi[left_index], yi[left_index]
        x2, y2 = xi[right_index], yi[right_index]

        # Subtract the excess from left interval before the left edge.
        # print(f" left {left} in {(x1, y1)}, {(x2, y2)}")
        if x1 < left:
            # Subtract the area of the rectangle from (x1, 0) to (left, y1)
            # plus 1/2 the rectangle from (x1, y1) to (left, y'),
            # where y' is y value where the line (x1, y1) to (x2, y2)
            # intersects x=left. This can be computed as follows:
            #    offset = left - x1
            #    slope = (y2 - y1)/(x2 - x1)
            #    yleft = y1 + slope*offset
            #    area = offset * y1 + offset * (yleft-y1)/2
            # It can be simplified to the following:
            #    area = offset * (y1 + slope*offset/2)
            offset = left - x1
            slope = (y2 - y1) / (x2 - x1)
            area = offset * (y1 + 0.5 * slope * offset)
            total -= area
            # print(f" left correction {area}")

        # Do trapezoidal integration up to and including the end interval
        while right_index < N_xi - 1 and x2 < right:
            # Add the current interval if it isn't empty
            if x1 != x2:
                area = 0.5 * (y1 + y2) * (x2 - x1)
                total += area
                # print(f" adding {(x1,y1)}, {(x2, y2)} as {area}")
            # Move to the next interval
            right_index += 1
            x1, y1, x2, y2 = x2, y2, xi[right_index], yi[right_index]
        if x1 != x2:
            area = 0.5 * (y1 + y2) * (x2 - x1)
            total += area
            # print(f" adding final {(x1,y1)}, {(x2, y2)} as {area}")

        # Subtract the excess from the right interval after the right edge.
        # print(f" right {right} in {(x1, y1)}, {(x2, y2)}")
        if x2 > right:
            # Expression for area to subtract using rectangles is as follows:
            #    offset = x2 - right
            #    slope = (y2 - y1)/(x2 - x1)
            #    yright = y2 - slope*offset
            #    area = -(offset * yright + offset * (y2-yright)/2)
            # It can be simplified to the following:
            #    area = -offset * (y2 - slope*offset/2)
            offset = x2 - right
            slope = (y2 - y1) / (x2 - x1)
            area = offset * (y2 - 0.5 * slope * offset)
            total -= area
            # print(f" right correction {area}")

        # Normalize by interval length
        if left < right:
            # print(f" normalize by length {right} - {left}")
            y[k] = total / (right - left)
        elif x1 < x2:
            # If dx = 0 using the value interpolated at x (with left=right=x).
            # print(f" dirac delta at {left} = {right} in {(x1, y1)}, {(x2, y2)}")
            offset = left - x1
            slope = (y2 - y1) / (x2 - x1)
            y[k] = y1 + slope * offset
        else:
            # At an empty interval in the theory function. Average the y.
            # print(f" empty interval with {left} = {right} in {(x1, y1)}, {(x2, y2)}")
            y[k] = 0.5 * (y1 + y2)


def convolve_gaussian_point(xin, yin, k, n, xo, limit, sigma):
    two_sigma_sq = 2.0 * sigma * sigma
    # double z, Glo, erflo, erfmin, y

    z = xo - xin[k]
    Glo = exp(-z * z / two_sigma_sq)
    erfmin = erflo = erf(-z / (SQRT2 * sigma))
    y = 0.0
    # /* printf("%5.3f: (%5.3f,%11.5g)",xo,xin[k],yin[k]); */
    while k < n - 1:
        k += 1
        if xin[k] != xin[k - 1]:
            # /* No additional contribution from duplicate points. */

            # /* Compute the next endpoint */
            zhi = xo - xin[k]
            Ghi = exp(-zhi * zhi / two_sigma_sq)
            erfhi = erf(-zhi / (SQRT2 * sigma))
            m = (yin[k] - yin[k - 1]) / (xin[k] - xin[k - 1])
            b = yin[k] - m * xin[k]

            # /* Add the integrals. */
            y += 0.5 * (m * xo + b) * (erfhi - erflo) - sigma / SQRT2PI * m * (Ghi - Glo)

            # /* Debug computation failures. */
            # if isnan(y) {
            #     print("NaN from %d: zhi=%g, Ghi=%g, erfhi=%g, m=%g, b=%g\n",
            #          % (k,zhi,Ghi,erfhi,m,b))
            # }

            # /* Save the endpoint for next trapezoid. */
            Glo = Ghi
            erflo = erfhi

            # /* Check if we've calculated far enough */
            if xin[k] >= xo + limit:
                break

    # /* printf(" (%5.3f,%11.5g)",xin[k<n?k:n-1],yin[k<n?k:n-1]); */

    # /* Normalize by the area of the truncated gaussian */
    # /* At this point erflo = erfmax */
    # /* printf ("---> %11.5g\n",2*y/(erflo-erfmin)); */
    return 2 * y / (erflo - erfmin)


def convolve_gaussian(xin, yin, x, dx, y):
    # size_t in,out;
    Nin = len(xin)
    Nout = len(x)

    # /* FIXME fails if xin are not sorted; slow if x not sorted */
    # assert(Nin>1)

    # /* Scan through all x values to be calculated */
    # /* Re: omp, each thread is going through the entire input array,
    # * independently, computing the resolution from the neighbourhood
    # * around its individual output points.  The firstprivate(in)
    # * clause sets each thread to keep its own copy of in, initialized
    # * at in's initial value of zero.  The "schedule(static,1)" clause
    # * puts neighbouring points in separate threads, which is a benefit
    # * since there will be less backtracking if resolution width increases
    # * from point to point.  Because the schedule is static, this does not
    # * significantly increase the parallelization overhead.  Because the
    # * threads are operating on interleaved points, there should be fewer cache
    # * misses than if each thread were given different stretches of x to
    # * convolve.
    # */
    k_in = 0

    for k_out in range(Nout):
        # /* width of resolution window for x is w = 2 dx^2. */
        sigma = dx[k_out]
        xo = x[k_out]
        limit = sqrt(-2.0 * sigma * sigma * LOG_RESLIMIT)

        # // if (out%20==0)

        # /* Line up the left edge of the convolution window */
        # /* It is probably forward from the current position, */
        # /* but if the next dx is a lot higher than the current */
        # /* dx or if the x are not sorted, then it may be before */
        # /* the current position. */
        # /* FIXME verify that the convolution window is just right */
        while k_in < Nin - 1 and xin[k_in] < xo - limit:
            k_in += 1
        while k_in > 0 and xin[k_in] > xo - limit:
            k_in -= 1

        # /* Special handling to avoid 0/0 for w=0. */
        if sigma > 0.0:
            y[k_out] = convolve_gaussian_point(xin, yin, k_in, Nin, xo, limit, sigma)
        elif k_in < Nin - 1:
            # /* Linear interpolation */
            m = (yin[k_in + 1] - yin[k_in]) / (xin[k_in + 1] - xin[k_in])
            b = yin[k_in] - m * xin[k_in]
            y[k_out] = m * xo + b
        elif k_in > 0:
            # /* Linear extrapolation */
            m = (yin[k_in] - yin[k_in - 1]) / (xin[k_in] - xin[k_in - 1])
            b = yin[k_in] - m * xin[k_in]
            y[k_out] = m * xo + b
        else:
            # /* Can't happen because there is more than one point in xin. */
            # assert(Nin>1)
            pass
